# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing XML functions"""

import io
import os
import tempfile

import pandas as pd

from ci.script_utils import files_path, xml_path
from sertit import ci, xml

ci.reduce_verbosity()


def _compare_xml_str(str_1: str, str_2: str) -> None:
    """
    Check XML converted to strings

    Args:
        str_1 (str): XML as string 1
        str_2 (str): XML as string 2
    """
    assert str_1.replace(" ", "").upper() == str_2.replace(" ", "").upper()


def _assert_str(str_1: str, str_2: str) -> None:
    """
    Check strings

    Args:
        str_1 (str): String 1
        str_2 (str): String 2
    """
    assert str_1 == str_2, f"str_1: {str_1} != str_2: {str_2}"


def test_xml():
    """Test XML functions"""

    # Check from pandas
    xml_str = """<?xml version='1.0' encoding='utf-8'?>
        <data>
         <row>
           <shape>square</shape>
           <degrees>360</degrees>
           <sides>4.0</sides>
         </row>
         <row>
           <shape>circle</shape>
           <degrees>360</degrees>
           <sides/>
         </row>
         <row>
           <shape>triangle</shape>
           <degrees>180</degrees>
           <sides>3.0</sides>
         </row>
        </data>
        """

    pd_xml = pd.read_xml(io.StringIO(xml_str))
    df_xml = xml.df_to_xml(pd_xml)
    _compare_xml_str(xml.to_string(df_xml), xml_str)

    # Check from dataset (using pandas is OK)
    pd_dict = {
        "Name": [
            "Owen Harris",
        ],
        "Age": [22],
        "Sex": ["male"],
    }
    pd_df = pd.DataFrame(pd_dict)

    attrs = list(pd_dict.keys())
    cv_xml = xml.convert_to_xml(pd_df, attrs)
    for attr in attrs:
        xml_attr = cv_xml.findtext(f".//{attr}")
        dict_attr = str(pd_dict[attr][0])
        assert xml_attr == dict_attr, f"{xml_attr=} =! {dict_attr=}"

    # Operations on XML
    xml.remove(cv_xml, "Age")
    assert cv_xml.findtext(".//Age") is None

    xml.add(cv_xml, "Age", 25)
    _assert_str(cv_xml.findtext(".//Age"), "25")

    xml.update_attrib(cv_xml, "Name", "prefix", "Mr.")
    _assert_str(cv_xml.find(".//Name").attrib["prefix"], "Mr.")

    xml.update_txt(cv_xml, "Age", 22)
    _assert_str(cv_xml.findtext(".//Age"), "22")

    # does nothing
    xml.update_txt(cv_xml, "feegr", 22)
    ci.assert_val(cv_xml.findtext(".//feegr"), None, "Does not exist")

    xml.update_txt_fct(cv_xml, "Age", lambda x: round(int(x) / 10) * 10)
    _assert_str(cv_xml.findtext(".//Age"), "20")

    # does nothing
    xml.update_txt_fct(cv_xml, "vczerhgrtz", lambda x: round(int(x) / 10) * 10)
    ci.assert_val(cv_xml.findtext(".//vczerhgrtz"), None, "Does not exist")

    # Write
    true_xml = str(xml_path() / "true.xml")
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_xml = os.path.join(tmp_dir, "tmp.xml")
        xml.write(cv_xml, tmp_xml)
        ci.assert_xml_equal(xml.read(true_xml), xml.read(tmp_xml))

    # Read archive
    # Based on `files.read_archived_xml`, so it is considered to work.
    # Just test the case with complete path to the archive
    l8_archived = files_path() / "LM05_L1TP_200030_20121230_20200820_02_T2_CI.zip"
    xml_archived = f"{l8_archived}!LM05_L1TP_200030_20121230_20200820_02_T2_CI/LM05_L1TP_200030_20121230_20200820_02_T2_MTL.xml"

    ci.assert_xml_equal(
        xml.read_archive(l8_archived, r".*_MTL\.xml"), xml.read_archive(xml_archived)
    )


def test_dict_xml():
    xml_dict = {
        "ACCODE": "LaSRC",
        "AREA_OR_POINT": "Area",
        "arop_ave_xshift(meters)": "0",
        "arop_ave_yshift(meters)": "0",
        "arop_ncp": "0",
        "arop_rmse(meters)": "0",
        "arop_s2_refimg": "NONE",
        "cloud_coverage": "46",
        "DATASTRIP_ID": "S2B_OPER_MSI_L1C_DS_2BPS_20221127T223847_S20221127T214906_N04.00",
        "HLS_PROCESSING_TIME": "2022-11-29T07:10:51Z",
        "HORIZONTAL_CS_CODE": "EPSG:32760",
        "HORIZONTAL_CS_NAME": "WGS84 / UTM zone 60S",
        "L1C_IMAGE_QUALITY": "NONE",
        "L1_PROCESSING_TIME": "2022-11-27T23:01:25.915427Z",
        "MEAN_SUN_AZIMUTH_ANGLE": "50.8114320532222",
        "MEAN_SUN_ZENITH_ANGLE": "37.2612285279726",
        "MEAN_VIEW_AZIMUTH_ANGLE": "104.448494899629",
        "MEAN_VIEW_ZENITH_ANGLE": "3.7539313224889",
        "MSI band 01 bandpass adjustment slope and offset": "0.995900,-0.000200",
        "NBAR_SOLAR_ZENITH": "38.4289484166757",
        "NCOLS": "3660",
        "NROWS": "3660",
        "OVR_RESAMPLING_ALG": "NEAREST",
        "PROCESSING_BASELINE": "04.00",
        "PRODUCT_URI": "S2B_MSIL1C_20221127T214909_N0400_R143_T60FXK_20221127T223847.SAFE",
        "SENSING_TIME": "2022-11-27T21:49:48.678784Z",
        "SPACECRAFT_NAME": "Sentinel-2B",
        "spatial_coverage": "53",
        "SPATIAL_RESOLUTION": "30",
        "TILE_ID": "S2B_OPER_MSI_L1C_TL_2BPS_20221127T223847_A029913_T60FXK_N04.00",
        "ULX": "600000",
        "ULY": "-5499960",
    }

    root = xml.dict_to_xml(xml_dict)

    for key, val in xml_dict.items():
        key_xml = key.replace(" ", "_").replace("(", "_").replace(")", "")
        _assert_str(root.findtext(f".//{key_xml}"), val)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing SNAP functions"""

import numpy as np
import pytest

from ci.script_utils import dask_env, display_path, s3_env
from sertit import ci, display, rasters_rio

ci.reduce_verbosity()


@s3_env
@dask_env()
def test_display():
    """Testing Display functions"""
    path_2d = display_path().joinpath("2d.tif")
    path_3d = display_path().joinpath("3d.tif")
    path_3d_minmax = display_path().joinpath("3d_minmax.tif")
    path_stack = display_path().joinpath("stack.tif")

    # Open data
    arr_2d, _ = rasters_rio.read(path_2d)
    arr_3d, _ = rasters_rio.read(path_3d)
    arr_3d_m, _ = rasters_rio.read(path_3d_minmax)
    stack, _ = rasters_rio.read(path_stack)

    # Scale data
    scaled_2d = display.scale(stack[0, ...])
    scaled_3d = display.scale(stack)
    scaled_3d_m = display.scale(stack, perc=0)

    # Test
    np.testing.assert_array_equal(scaled_2d, arr_2d[0, ...])
    np.testing.assert_almost_equal(scaled_3d, arr_3d)
    np.testing.assert_almost_equal(scaled_3d_m, arr_3d_m)  # Almost here

    # Scale to uint8
    scaled_uint8_2d = display.scale_to_uint8(stack[0, ...])
    scaled_uint8_3d = display.scale_to_uint8(stack)

    assert scaled_uint8_2d.dtype == scaled_uint8_3d.dtype == np.uint8
    np.testing.assert_array_equal(
        scaled_uint8_2d, (arr_2d[0, ...] * 254 + 1).astype(np.uint8)
    )
    np.testing.assert_array_equal(scaled_uint8_3d, (arr_3d * 254 + 1).astype(np.uint8))

    with pytest.raises(ValueError):
        display.scale(np.expand_dims(stack, 0))



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the logs"""

import logging
import os
import sys
import tempfile

import colorlog as clog
import pytest

from sertit import ci, logs
from sertit.logs import LOGGING_FORMAT

ci.reduce_verbosity()


def test_log():
    """Testing log functions"""

    # -- INIT LOGGER --
    log_lvl = logging.WARNING

    logger = logging.getLogger("Test_logger")
    logger.handlers.append(logging.StreamHandler())  # Handler to remove
    logs.init_logger(logger, log_lvl=log_lvl, log_format=LOGGING_FORMAT)
    logs.reset_logging()
    assert logger.handlers == []
    assert logger.filters == []
    assert logger.level == 0

    # Re init logger
    logs.init_logger(logger, log_lvl=log_lvl, log_format=LOGGING_FORMAT)

    # Test init
    assert logger.level == log_lvl
    assert logger.handlers[0].formatter._fmt == LOGGING_FORMAT

    # -- CREATE LOGGER --
    logger_test = logging.getLogger("test")
    file_log_lvl = logging.DEBUG
    stream_log_lvl = logging.DEBUG
    with tempfile.TemporaryDirectory() as tmp_dir:
        logs.create_logger(
            logger,
            file_log_level=file_log_lvl,
            stream_log_level=stream_log_lvl,
            output_folder=tmp_dir,
            name="test_log.txt",
            other_loggers_names=["test"],
        )

        logger.info("Hey you!")

        # Test create
        assert len(logger.handlers) == 2  # File and stream
        assert len(logger_test.handlers) == 2  # File and stream
        for handler in logger.handlers + logger_test.handlers:
            if isinstance(handler, logging.FileHandler):
                assert handler.level == file_log_lvl
                log_file = handler.baseFilename
                assert log_file is not None and os.path.isfile(log_file)
            elif isinstance(handler, logging.StreamHandler):
                assert handler.level == stream_log_lvl
            else:
                raise TypeError(
                    f"Invalid handler type: {handler.__class__}, "
                    f"the logger should only have Stream and File handlers"
                )

        logs.reset_logging()
        logs.create_logger(
            logger, stream_log_level=stream_log_lvl, other_loggers_names="test"
        )

        logger.info("Hi there!")

        # Test create
        assert len(logger.handlers) == 1  # Only stream
        assert len(logger_test.handlers) == 1  # Only stream
        colored_fmt_cls = clog.ColoredFormatter
        for handler in logger.handlers + logger_test.handlers:
            if isinstance(handler, logging.StreamHandler):
                assert handler.level == stream_log_lvl
                assert isinstance(handler.formatter, colored_fmt_cls)
            else:
                raise TypeError(
                    f"Invalid handler type: {handler.__class__}, "
                    f"the logger should only have Stream handler"
                )

        # Without color
        import colorlog

        colorlog_sys = sys.modules["colorlog"]
        del colorlog
        sys.modules["colorlog"] = None
        logs.create_logger(logger, stream_log_level=stream_log_lvl)

        logger.info("Urk!")

        for handler in logger.handlers:
            assert isinstance(handler.formatter, logging.Formatter)

        # Just in case
        sys.modules["colorlog"] = colorlog_sys

        # Cleanup
        logs.shutdown_logger(logger)
        logs.shutdown_logger(logger_test)

        logs.reset_logging()

        # Reduce verbosity again after resetting
        ci.reduce_verbosity()


def test_deprecation_warning():
    # Just test this doesn't throw an error
    with pytest.deprecated_call():
        logs.deprecation_warning("This is deprecated.")



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the CI"""

import logging
import os

import geopandas
import pytest
import rasterio
from cloudpathlib import AnyPath, S3Client
from tempenv import tempenv

from ci.script_utils import CI_SERTIT_S3
from sertit import ci, rasters, vectors
from sertit.ci import AWS_S3_ENDPOINT
from sertit.logs import SU_NAME
from sertit.s3 import USE_S3_STORAGE, s3_env, temp_s3

LOGGER = logging.getLogger(SU_NAME)


def base_fct(value):
    raster_path = AnyPath("s3://sertit-ci/sertit_utils").joinpath(
        "DATA", "rasters", "raster.tif"
    )
    assert raster_path.client.client.meta.endpoint_url == "https://s3.unistra.fr"
    assert raster_path.is_file()
    assert rasters.read(raster_path).rio.count == 1
    return 1


@s3_env(use_s3_env_var=CI_SERTIT_S3)
def with_s3(variable_1, variable_2):
    return base_fct(1)


def without_s3():
    S3Client().set_as_default_client()
    return base_fct(None)


def test_s3_raster():
    with tempenv.TemporaryEnvironment(
        {USE_S3_STORAGE: "1", AWS_S3_ENDPOINT: "s3.unistra.fr", CI_SERTIT_S3: "1"}
    ):
        # There is a mistake in endpoint but AWS_S3_ENDPOINT should override it
        with temp_s3(endpoint="mistake.s3.unistra.fr"):
            raster_path = AnyPath("s3://sertit-ci/sertit_utils").joinpath(
                "DATA", "rasters", "raster.tif"
            )
            assert (
                raster_path.client.client.meta.endpoint_url == "https://s3.unistra.fr"
            )
            assert raster_path.is_file()

            # Test only rasterio since rasters module may interfere
            assert rasterio.open(str(raster_path)).count == 1
            assert rasters.read(raster_path).rio.count == 1

        with pytest.raises(AssertionError):
            without_s3()
        assert with_s3(1, 2) == 1


def test_s3_vector():
    # There is a mistake in endpoint but AWS_S3_ENDPOINT should override it
    with (
        tempenv.TemporaryEnvironment({USE_S3_STORAGE: "1", CI_SERTIT_S3: "1"}),
        temp_s3(endpoint="s3.unistra.fr"),
    ):
        vector_path = AnyPath("s3://sertit-ci/sertit_utils").joinpath(
            "DATA", "vectors", "aoi.shp"
        )
        assert vector_path.client.client.meta.endpoint_url == "https://s3.unistra.fr"
        assert vector_path.is_file()
        # Test geopandas since vectors module could interfere with s3 configuration
        assert geopandas.read_file(str(vector_path)).shape[0] == 1
        assert vectors.read(vector_path).shape[0] == 1


def test_no_sign_request():
    with (
        tempenv.TemporaryEnvironment(
            {
                "AWS_S3_ENDPOINT": "s3.us-west-2.amazonaws.com",
            }
        ),
        temp_s3(no_sign_request=True),
    ):
        path = AnyPath(
            "s3://sentinel-cogs/sentinel-s2-l2a-cogs/40/V/DR/2023/11/S2A_40VDR_20231114_0_L2A"
        )
        assert path.exists()
        with rasterio.open(str(path / "B12.tif")) as ds:
            assert ds.meta["dtype"] == "uint16"


def test_requester_pays():
    with (
        tempenv.TemporaryEnvironment(
            {
                "AWS_S3_ENDPOINT": "s3.eu-central-1.amazonaws.com",
                "AWS_SECRET_ACCESS_KEY": os.getenv("AWS_S3_AWS_SECRET_ACCESS_KEY"),
                "AWS_ACCESS_KEY_ID": os.getenv("AWS_S3_AWS_ACCESS_KEY_ID"),
            }
        ),
        temp_s3(requester_pays=True),
    ):
        path = AnyPath("s3://sentinel-s2-l1c/tiles/29/H/NA/2023/11/14/0/")
        assert path.exists()
        with rasterio.open(str(path / "B12.jp2")) as ds:
            assert ds.meta["dtype"] == "uint16"


def ko_to_bytes(value):
    return int(value * 1e3)


def mo_to_bytes(value):
    return int(value * 1e6)


def _update_s3_env(function):
    def s3_env_wrapper(*args, **kwargs):
        from sertit import unistra

        LOGGER.debug("In decorator!")
        with rasterio.Env(
            # Set to TRUE or EMPTY_DIR to avoid listing all files in the directory once a single file is opened (this is highly recommended).
            GDAL_DISABLE_READDIR_ON_OPEN=True,
            # Size of the default block cache, can be set in byte, MB, or as a percentage of available main, memory.
            GDAL_CACHEMAX=mo_to_bytes(200),
            # Global cache size for downloads in bytes, defaults to 16 MB.
            CPL_VSIL_CURL_CACHE_SIZE=mo_to_bytes(200),
            # Enable / disable per-file caching by setting to TRUE or FALSE.
            VSI_CACHE=True,
            # Per-file cache size in bytes
            VSI_CACHE_SIZE=mo_to_bytes(5),
            # When set to YES, this attempts to download multiple range requests in parallel, reusing the same TCP connection
            GDAL_HTTP_MULTIPLEX=True,
            # Gives the number of initial bytes GDAL should read when opening a file and inspecting its metadata.
            GDAL_INGESTED_BYTES_AT_OPEN=ko_to_bytes(32),
            GDAL_HTTP_VERSION=2,
            # Tells GDAL to merge consecutive range GET requests.
            GDAL_HTTP_MERGE_CONSECUTIVE_RANGES="YES",
            # Number of threads GDAL can use for block reads and (de)compression, set to ALL_CPUS to use all available cores.
            GDAL_NUM_THREADS="ALL_CPUS",
        ):
            return unistra.s3_env(function(*args, **kwargs), *args, **kwargs)

    return s3_env_wrapper


def _fct():
    LOGGER.debug("In _fct!")
    assert rasterio.env.hasenv()
    rasterio_env = rasterio.env.getenv()
    LOGGER.info(rasterio_env)
    print(rasterio_env)
    assert "GDAL_CACHEMAX" in rasterio_env
    assert "VSI_CACHE" in rasterio_env
    assert "GDAL_HTTP_VERSION" in rasterio_env
    assert "GDAL_DISABLE_READDIR_ON_OPEN" in rasterio_env
    ci.assert_val(rasterio_env["GDAL_CACHEMAX"], mo_to_bytes(200), "GDAL_CACHEMAX")
    ci.assert_val(rasterio_env["VSI_CACHE"], True, "VSI_CACHE")
    ci.assert_val(rasterio_env["GDAL_HTTP_VERSION"], 2, "GDAL_HTTP_VERSION")
    ci.assert_val(rasterio_env["GDAL_DISABLE_READDIR_ON_OPEN"], True, "VSI_CACHE")


@_update_s3_env
def _decorated_fct():
    return _fct()


def test_rasterio_env():
    # See https://developmentseed.org/titiler/advanced/performance_tuning/#recommended-configuration-for-dynamic-tiling
    # And https://gdalcubes.github.io/source/concepts/config.html#recommended-settings-for-cloud-access

    _decorated_fct()

    with pytest.raises(AssertionError):
        _fct()



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Vectors tools

You can use this only if you have installed sertit[full] or sertit[vectors]
"""

import contextlib
import logging
import os
import re
import shutil
import tarfile
import tempfile
import zipfile
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

import geopandas as gpd
import numpy as np
import pandas as pd
from cloudpathlib.exceptions import AnyPathTypeError
from shapely import Polygon, wkt
from shapely.geometry.base import BaseGeometry

from sertit import AnyPath, files, geometry, misc, path, strings
from sertit.logs import SU_NAME
from sertit.types import AnyPathStrType, AnyPathType

LOGGER = logging.getLogger(SU_NAME)

EPSG_4326 = "EPSG:4326"
WGS84 = EPSG_4326

EXT_TO_DRIVER = {
    ".shp": "ESRI Shapefile",
    ".kml": "KML",
    ".kmz": "KMZ",
    ".json": "GeoJSON",
    ".geojson": "GeoJSON",
    ".gml": "GML",
}

SHP_CO_FILES = [".dbf", ".prj", ".sbn", ".sbx", ".shx", ".sld"]


def is_geopandas_1_0():
    """Is geopandas over 1.0.0. Default engine changes, from fiona to pyogrio"""
    return misc.compare_version("geopandas", "1.0.0", ">=")


if is_geopandas_1_0():
    from pyogrio.errors import DataSourceError
    from shapely.errors import GEOSException

    CPLE_AppDefinedError = Exception
else:  # pragma: no cover
    from fiona._err import CPLE_AppDefinedError
    from fiona.errors import DriverError as DataSourceError
    from fiona.errors import UnsupportedGeometryTypeError as GEOSException

# Handle errors changing from fiona to pyogrio
DataSourceError = DataSourceError
GEOSException = GEOSException
CPLE_AppDefinedError = CPLE_AppDefinedError


def to_utm_crs(lon: float, lat: float) -> "CRS":  # noqa: F821
    """
    Find the EPSG code of the UTM CRS from a lon/lat in WGS84.

    Args:
        lon (float): Longitude (WGS84, epsg:4326)
        lat (float): Latitude (WGS84, epsg:4326)

    Returns:
        CRS: UTM CRS

    Example:
        >>> to_utm_crs(lon=7.8, lat=48.6)  # Strasbourg
        <Derived Projected CRS: EPSG:32632>
        Name: WGS 84 / UTM zone 32N
        Axis Info [cartesian]:
        - E[east]: Easting (metre)
        - N[north]: Northing (metre)
        Area of Use:
        - bounds: (6.0, 0.0, 12.0, 84.0)
        Coordinate Operation:
        - name: UTM zone 32N
        - method: Transverse Mercator
        Datum: World Geodetic System 1984 ensemble
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich

    """
    # Manage the case with centroids etc. that are already written as arrays
    try:
        point = gpd.points_from_xy([lon], [lat])
    except ValueError:
        point = gpd.points_from_xy(lon, lat)

    return gpd.GeoDataFrame(geometry=point, crs=EPSG_4326).estimate_utm_crs()


def get_geodf(geom: BaseGeometry | list | gpd.GeoSeries, crs: str) -> gpd.GeoDataFrame:
    """
    Get a GeoDataFrame from a geometry and a crs

    Args:
        geom (BaseGeometry | list | gpd.GeoSeries): List of Polygons, or Polygon or bounds
        crs (str): CRS of the polygon

    Returns:
        gpd.GeoDataFrame: Geometry as a geodataframe

    Example:
        >>> poly = Polygon(((0., 0.), (0., 1.), (1., 1.), (1., 0.), (0., 0.)))
        >>> geodf = get_geodf(poly, crs=WGS84)
        >>> print(geodf)
                                                    geometry
        0  POLYGON ((0.00000 0.00000, 0.00000 1.00000, 1....
    """
    if isinstance(geom, list):
        if isinstance(geom[0], BaseGeometry):
            pass
        else:
            try:
                geom = [geometry.from_bounds_to_polygon(*geom)]
            except TypeError as ex:
                raise TypeError(
                    "Give the extent as 'left', 'bottom', 'right', and 'top'"
                ) from ex
    elif isinstance(geom, BaseGeometry):
        geom = [geom]
    elif isinstance(geom, gpd.GeoSeries):
        geom = geom.geometry
    else:
        raise TypeError("geometry should be a list or a Geometry.")

    return gpd.GeoDataFrame(geometry=geom, crs=crs)


def set_kml_driver() -> None:
    """
    Set KML driver for Fiona data (use it at your own risks !)

    Only useful with :code:`geopandas<1.0.0` or with :code:`fiona`'s engine

    Example:
        >>> path = "path/to/kml.kml"
        >>> gpd.read_file(path, engine="fiona")
        fiona.errors.DriverError: unsupported driver: 'LIBKML'
        >>> set_kml_driver()
        >>> gpd.read_file(path, engine="fiona")
                       Name  ...                                           geometry
        0  CC679_new_AOI2_3  ...  POLYGON Z ((45.03532 32.49765 0.00000, 46.1947...
        [1 rows x 12 columns]

    """
    if not is_geopandas_1_0():  # pragma: no cover
        import fiona

        drivers = fiona.drvsupport.supported_drivers

        if "LIBKML" not in drivers:
            drivers["LIBKML"] = "rw"
        if "KML" not in drivers:  # Just in case
            drivers["KML"] = "rw"


def get_aoi_wkt(aoi_path: AnyPathStrType, as_str: bool = True) -> tuple[str | Polygon]:
    """
    Get AOI formatted as a WKT from files that can be read by Fiona (like shapefiles, ...)
    or directly from a WKT file. The use of KML has been forced (use it at your own risks !).

    See: https://fiona.readthedocs.io/en/latest/fiona.html#fiona.open

    It is assessed that:

    - only **one** polygon composes the AOI (as only the first one is read)
    - it should be specified in lat/lon (WGS84) if a WKT file is provided

    Args:
        aoi_path (AnyPathStrType): Absolute or relative path to an AOI.
            Its format should be WKT or any format read by Fiona, like shapefiles.
        as_str (bool): If True, return WKT as a str, otherwise as a shapely geometry

    Returns:
         str | Polygon: AOI formatted as a WKT stored in lat/lon

    Example:
        >>> path = "path/to/vec.geojson"  # OK with ESRI Shapefile, geojson, WKT, KML...
        >>> get_aoi_wkt(path)
        'POLYGON Z ((46.1947755465253067 32.4973553439109324 0.0000000000000000, 45.0353174370802520 32.4976496856158974
        0.0000000000000000, 45.0355748149750283 34.1139970085580018 0.0000000000000000, 46.1956059695554089
        34.1144793800670882 0.0000000000000000, 46.1947755465253067 32.4973553439109324 0.0000000000000000))'


    """
    aoi_path = AnyPath(aoi_path)
    if not aoi_path.is_file():
        raise FileNotFoundError(f"AOI file {aoi_path} does not exist.")

    if aoi_path.suffix == ".wkt":
        try:
            with open(aoi_path) as aoi_f:
                aoi = wkt.load(aoi_f)
        except Exception as ex:  # pragma: no cover
            raise ValueError("AOI WKT cannot be read") from ex
    else:
        try:
            # Open file
            aoi_file = read(aoi_path, crs=EPSG_4326)

            # Get envelope polygon
            geom = aoi_file["geometry"]
            if len(geom) > 1:  # pragma: no cover
                LOGGER.warning(
                    "Your AOI contains several polygons. Only the first will be treated !"
                )
            polygon = geom[0].convex_hull

            # Convert to WKT
            aoi = wkt.loads(str(polygon))

        except Exception as ex:  # pragma: no cover
            raise ValueError("AOI cannot be read by Fiona") from ex

    # Convert to string if needed
    if as_str:
        aoi = wkt.dumps(aoi)

    LOGGER.debug("Specified AOI in WKT: %s", aoi)
    return aoi


def shapes_to_gdf(shapes: Generator, crs: str) -> gpd.GeoDataFrame:
    """
    Convert rasterio shapes to geodataframe

    Args:
        shapes (Generator): Shapes from rasterio
        crs: Wanted CRS of the vector. If None, using naive or origin CRS.

    Returns:
        gpd.GeoDataFrame: Shapes as a GeoDataFrame
    """

    def _to_polygons(val: Any) -> Polygon:
        """
        Convert to polygon (to be used in pandas) -> convert the geometry column

        Args:
            val (Any): Pandas value that has a "coordinates" field

        Returns:
            Polygon: Pandas value as a Polygon
        """
        # Donut cases
        if len(val["coordinates"]) > 1:
            poly = Polygon(val["coordinates"][0], val["coordinates"][1:])
        else:
            poly = Polygon(val["coordinates"][0])
            pass

        return poly

    # Convert results to pandas (because of invalid geometries) and save it
    pd_results = pd.DataFrame(shapes, columns=["geometry", "raster_val"])

    if not pd_results.empty:
        # Convert to proper polygons(correct geometries)
        pd_results.geometry = pd_results.geometry.apply(_to_polygons)

    # Convert to geodataframe with correct geometry
    gdf = gpd.GeoDataFrame(pd_results, geometry=pd_results.geometry, crs=crs)

    # Return valid geometries
    gdf = geometry.make_valid(gdf)

    return gdf


def write(gdf: gpd.GeoDataFrame, path: AnyPathStrType, **kwargs) -> None:
    """
    Write vector to disk, managing the common drivers automatically.

    Args:
        gdf (gpd.GeoDataFrame): GeoDataFrame to write on disk
        path (AnyPathStrType): Where to write on disk.

    Example:
        >>> write(my_vector, path="my_vector.kml")
    """
    path = AnyPath(path)

    driver = kwargs.pop("driver", None)
    if not driver:
        driver = EXT_TO_DRIVER.get(path.suffix)
        if driver == "KML":
            set_kml_driver()
        elif driver == "KMZ":
            raise NotImplementedError("Impossible to write a KMZ for now.")

    gdf.to_file(str(path), driver=driver, **misc.prune_dict(kwargs, ["window"]))


def copy(src_path: AnyPathStrType, dst_path: AnyPathStrType) -> AnyPathType:
    """
     Copy vector (handles shapefiles additional files)

    Args:
        src_path (AnyPathStrType): Source Path
        dst_path (AnyPathStrType): Destination Path (file or folder)

    Returns:
        AnyPathType: Path to copied vector

    Example:
        >>> new_aoi_path = copy("in/aoi.shp", "out/aoi.shp")
        >>> new_aoi_path
        PosixPath('out/aoi.shp')
        >>> list(new_aoi_path.parent.glob("*"))
        [PosixPath('out/aoi.dbf'), PosixPath('out/aoi.prj'), PosixPath('out/aoi.shp'), PosixPath('out/aoi.shx')]
    """
    src_path = AnyPath(src_path)
    dst_path = AnyPath(dst_path)

    if not dst_path.is_file():
        dst_path = files.copy(src_path, dst_path)

        # Add files that come with shape
        shp_co_files = [
            file
            for file in src_path.parent.glob(f"{path.get_filename(src_path)}.*")
            if file.suffix in SHP_CO_FILES
        ]
        for co_file in shp_co_files:
            files.copy(co_file, dst_path.with_suffix(co_file.suffix))

    return dst_path


def read(
    vector_path: AnyPathStrType,
    crs: Any = None,
    archive_regex: str = None,
    window: Any = None,
    **kwargs,
) -> gpd.GeoDataFrame:
    """
    Read any vector:

    - if KML/KMZ: sets correctly the drivers and open layered KML (you may need :code:`ogr2ogr` to make it work !)
    - if archive (only zip or tar), use a regex to look for the vector inside the archive. You can use this `site <https://regexr.com/>`_ to build your regex.
    - if GML: manages the empty errors

    Handles a lot of exceptions and have fallback mechanisms with :code:`ogr2ogr` (if in :code:`PATH`)

    Handles both on disk and cloud-stored vectors.

    Adds two attributes to your vector:

    - :code:`path`: File path
    - :code:`name`: File name

    Args:
        vector_path (AnyPathStrType): Path to vector to read. In case of archive, path to the archive.
        crs: Wanted CRS of the vector. If None, using naive or origin CRS.
        archive_regex (str): [Archive only] Regex for the wanted vector inside the archive
        window (Any): Anything that can be returned as a bbox (i.e. path, gpd.GeoPandas, Iterable, ...).
            In case of an iterable, assumption is made it corresponds to geographic bounds. Mimics :code:`rasters.read(..., window=)`. If given, :code:`bbox` is ignored.
        **kwargs: Additional arguments used in gpd.read_file.
            You can also give :code:`file_list`, the list of files of the archive to get the vector from, as this operation is expensive when done with large archives stored on the cloud.
            You can also set :code:`compute_sindex=False` to avoid computing the spatial index of the vector.

    Returns:
        gpd.GeoDataFrame: Read vector as a GeoDataFrame

    Examples:
        >>> # Usual
        >>> path = 'D:/path/to/vector.geojson'
        >>> vec = vectors.read(path, crs=WGS84)
                               Name  ...                                           geometry
        0  Sentinel-1 Image Overlay  ...  POLYGON ((0.85336 42.24660, -2.32032 42.65493,...
        >>>
        >>> # Attributes
        >>> vec.attrs["path"]
        'D:/path/to/vector.geojson'
        >>>
        >>> vec.attrs["name"]
        'vector.geojson'

        >>> # Archive
        >>> arch_path = 'D:/path/to/zip.zip'
        >>> vectors.read(arch_path, archive_regex=r".*map-overlay.kml")
                               Name  ...                                           geometry
        0  Sentinel-1 Image Overlay  ...  POLYGON ((0.85336 42.24660, -2.32032 42.65493,...
    """
    # Default values
    gpd_vect_path = str(vector_path)
    arch_path = None

    # -- Manage window and convert it to a bbox
    if window is not None:
        try:
            bbox = read(window)
        except (FileNotFoundError, TypeError):
            # Convert ndarray to tuple
            bbox = tuple(window) if isinstance(window, np.ndarray) else window

        kwargs["bbox"] = bbox

    # -- Manage the path formatting (create the path to be read by GeoPandas, the archive path if needed, ...)
    try:
        vector_path = AnyPath(vector_path)

        # Manage formatted archive file (fsspec style for example)
        if "!" in str(vector_path):
            split_vect = str(vector_path).split("!")
            archive_regex = ".*{}".format(split_vect[1].replace(".", r"\."))
            vector_path = AnyPath(split_vect[0])

        # Manage archive case
        if vector_path.suffix in [".tar", ".zip"]:
            prefix = vector_path.suffix[-3:]
            file_list = kwargs.pop(
                "file_list", path.get_archived_file_list(vector_path)
            )

            try:
                regex = re.compile(archive_regex)
                arch_path = list(filter(regex.match, file_list))[0]

                # Different template if on cloud or not... (only tested with S3)
                if path.is_cloud_path(vector_path):
                    gpd_vect_path = f"{prefix}+{vector_path}!{arch_path}"
                else:
                    gpd_vect_path = f"{prefix}://{vector_path}!{arch_path}"
            except IndexError as exc:
                raise FileNotFoundError(
                    f"Impossible to find vector {archive_regex} in {path.get_filename(vector_path)}"
                ) from exc
        # Don't read tar.gz archives (way too slow)
        elif vector_path.suffixes == [".tar", ".gz"]:  # pragma: no cover
            raise TypeError(
                ".tar.gz files are too slow to be read from inside the archive. Please extract them instead."
            )
    except AnyPathTypeError:
        pass

    # Check existence of the file (here and not before to handle fsspec cases with '!')
    if not AnyPath(vector_path).exists():
        raise FileNotFoundError(f"Non existing vector: {vector_path}")

    # Read vector
    vect = _read_vector_core(gpd_vect_path, vector_path, arch_path, crs, **kwargs)

    # Add some attributes
    vect.attrs["path"] = str(vector_path)
    vect.attrs["name"] = path.get_filename(vector_path)

    # Generate spatial index for optimization
    with contextlib.suppress(AttributeError):
        if kwargs.get("compute_sindex", True) and not vect.has_sindex:
            vect.sindex  # noqa

    return vect


def _read_vector_core(
    gpd_vect_path: str, raw_path: AnyPathStrType, arch_path: str, crs, **kwargs
):
    """
    Read vector (core function) with correctly formatted paths.

    Handles a lot of exceptions, reads KML, and have fallback mechanisms with ogr2ogr (if in PATH)

    Args:
        gpd_vect_path (str): Resolved vector path (readable by geopandas)
        raw_path (AnyPathStrType): Path to vector to read. In case of archive, path to the archive.
        arch_path (str): If archived vector, path to the vector file inside the archive (from the root of the archive)
        crs: Wanted CRS of the vector. If None, using naive or origin CRS.
        **kwargs: Other arguments

    Returns:
        gpd.GeoDataFrame: Read vector as a GeoDataFrame
    """
    tmp_dir = None

    # -- Open vector
    try:
        # Discard some weird error concerning a NULL pointer that outputs a ValueError (as we already except it)
        fiona_logger = logging.getLogger("fiona")
        fiona_logger.setLevel(logging.CRITICAL)

        # Manage KML driver
        if gpd_vect_path.endswith(".kml") or gpd_vect_path.endswith(".kmz"):
            vect = _read_kml(gpd_vect_path, raw_path, arch_path, tmp_dir, **kwargs)
        else:
            vect = gpd.read_file(gpd_vect_path, **kwargs)

        # Manage naive geometries
        try:
            if vect.crs and crs:
                vect = vect.to_crs(crs)
        except AttributeError:
            # Pyogrio don't create crs columns for dbf files for example
            pass

        # Set fiona logger back to what it was
        fiona_logger.setLevel(logging.INFO)
    except DataSourceError:  # pragma: no cover
        raise
    except (ValueError, GEOSException, IndexError) as ex:
        if "Use a.any() or a.all()" in str(ex):
            raise
        # Do not print warning for null layer
        elif "Null layer" not in str(ex):
            LOGGER.warning(ex)
        vect = gpd.GeoDataFrame(geometry=[], crs=crs)
    except CPLE_AppDefinedError as ex:
        # CPLE_AppDefinedError is not a pyogrio exception and this is therefore too broad
        if is_geopandas_1_0() and kwargs.get("engine") != "fiona":
            raise ex

        # Last try to read this vector
        # Needs ogr2ogr here
        if shutil.which("ogr2ogr"):  # pragma: no cover
            # Open as geojson
            tmp_dir = tempfile.TemporaryDirectory()
            vect_path_gj = ogr2geojson(raw_path, tmp_dir.name, arch_path)
            vect = gpd.read_file(vect_path_gj, **kwargs)
            vect.crs = None
        else:  # pragma: no cover
            # Do not print warning for null layer
            if "Null layer" not in str(ex):
                LOGGER.warning(ex)
            vect = gpd.GeoDataFrame(geometry=[], crs=crs)

    # Clean if needed
    if tmp_dir:  # pragma: no cover
        tmp_dir.cleanup()

    return vect


def _read_kml(
    gpd_vect_path: str,
    raw_path: AnyPathStrType,
    arch_path: str = None,
    tmp_dir=None,
    **kwargs,
) -> gpd.GeoDataFrame:
    """
    Reader of KML data

    Args:
        gpd_vect_path (str): Resolved vector path (readable by geopandas)
        raw_path (AnyPathStrType): Path to vector to read. In case of archive, path to the archive.
        arch_path: If archived vector, path to the vector file inside the archive (from the root of the archive)
        tmp_dir: Temporary directory
        **kwargs: Additional arguments used in gpd.read_file

    Returns:
        gpd.GeoDataFrame: KML as a geopandas GeoDataFrame

    """
    vect = gpd.GeoDataFrame()
    driver = "KML" if gpd_vect_path.endswith(".kml") else "KMZ"
    engine = None

    # Errors reading KML and KMZ with pyogrio until v0.12.0
    # https://github.com/geopandas/pyogrio/issues/543
    # https://github.com/geopandas/pyogrio/issues/444
    use_pyogrio = is_geopandas_1_0()
    from importlib.metadata import version

    if misc.compare_version("pyogrio", "0.12.0", "<"):
        engine = "fiona"
        use_pyogrio = False

    if use_pyogrio:
        try:
            vect = gpd.read_file(gpd_vect_path, driver=driver, engine=engine, **kwargs)
        except DataSourceError:  # pragma: no cover
            LOGGER.error(
                f"Error in reading {path.get_filename(gpd_vect_path)}. geopandas: {version('geopandas')},  pyogrio: {version('pyogrio')}, engine: {engine}"
            )
    else:
        import fiona

        set_kml_driver()

        # Document tags in KML file are separate layers for GeoPandas.
        # When you try to get the KML content, you actually get the first layer.
        # So you need for loop for iterating over layers.
        # https://gis.stackexchange.com/questions/328525/geopandas-read-file-only-reading-first-part-of-kml/328554
        for layer in fiona.listlayers(gpd_vect_path):
            try:
                vect_layer = gpd.read_file(
                    gpd_vect_path, driver=driver, layer=layer, engine=engine, **kwargs
                )
                if not vect_layer.empty:  # pragma: no cover
                    # KML files are always in WGS84 (and does not contain this information)
                    vect_layer.crs = EPSG_4326
                    vect = pd.concat([vect, vect_layer])
            except ValueError:
                pass  # Except Null Layer

    # Workaround for archived KML -> they may be empty
    # Convert KML to GeoJSON
    # Needs ogr2ogr here
    if vect.empty:
        if shutil.which("ogr2ogr"):
            LOGGER.debug(
                "Impossible to open your KML file with Python. Using 'ogr2ogr' to convert it into a more readable format."
            )
            # Open the geojson
            if not tmp_dir:
                tmp_dir = tempfile.TemporaryDirectory()

            vect_path_gj = ogr2geojson(raw_path, tmp_dir.name, arch_path)
            vect = gpd.read_file(vect_path_gj, **kwargs)
        else:  # pragma: no cover
            # Try reading it in a basic manner
            LOGGER.warning(
                "Missing `ogr2ogr` in your PATH, your KML may be incomplete. "
                "(KML files can contain unsupported data structures, nested folders etc.)"
            )
            try:
                vect = gpd.read_file(gpd_vect_path, **kwargs)
            except Exception:
                # Force set CRS to empty vector
                vect.crs = EPSG_4326

    return vect


def ogr2geojson(
    vector_path: AnyPathStrType,
    out_dir: AnyPathStrType,
    arch_vect_path: str = None,
) -> str:
    """
    Wrapper of ogr2ogr function, converting the input vector to GeoJSON.

    Args:
        vector_path (AnyPathStrType): Path to vector to read. In case of archive, path to the archive.
        out_dir (AnyPathStrType): Output directory
        arch_vect_path: If archived vector, path to the vector file inside the archive (from the root of the archive)

    Returns:
        str: Converted file

    Args:
        vector_path (AnyPathStrType): Path to vector to read. In case of archive, path to the archive.
        out_dir (AnyPathStrType): Output directory
        arch_vect_path: If archived vector, path to the vector file inside the archive (from the root of the archive)

    Returns:
        str: Converted file
    """
    assert shutil.which("ogr2ogr")  # Needs ogr2ogr here

    # Convert to strings to make it work with CLI
    vector_path = AnyPath(vector_path)

    # archived vector_path are extracted in a tmp folder so no need to be downloaded
    if vector_path.suffix == ".zip":
        with zipfile.ZipFile(vector_path, "r") as zip_ds:
            vect_path = zip_ds.extract(arch_vect_path, out_dir)
    elif vector_path.suffix == ".tar":
        with tarfile.open(vector_path, "r") as tar_ds:
            tar_ds.extract(arch_vect_path, out_dir)
            vect_path = os.path.join(out_dir, arch_vect_path)
    else:
        # vector_path should be downloaded to work with 'ogr2ogr'
        if path.is_cloud_path(vector_path):
            vector_path = AnyPath(vector_path).fspath
        vect_path = vector_path

    vect_path_gj = os.path.join(
        out_dir,
        os.path.basename(vect_path).replace(path.get_ext(vect_path), "geojson"),
    )
    cmd_line = [
        "ogr2ogr",
        "-fieldTypeToString DateTime",  # Disable warning
        "-f GeoJSON",
        strings.to_cmd_string(vect_path_gj),  # dst
        strings.to_cmd_string(vect_path),  # src
    ]
    try:
        misc.run_cli(cmd_line)
    except RuntimeError as ex:  # pragma: no cover
        raise RuntimeError(f"Something went wrong with ogr2ogr: {ex}") from ex

    return vect_path_gj


@contextmanager
def utm_crs(gdf: gpd.GeoDataFrame) -> None:
    """
    Change temporary the CRS of a vector, ie when computing area based statistics / features (centroid....) which need a meter-based CRS.

    WARNING:
        The modifications (other than CRS) on the yielded GeoDataFrame will be kept!

    Args:
        gdf (str): GeoDataFrame to convert

    Example:
        >>> vect = vectors.read(vectors_path().joinpath("aoi.kml"))
        >>> with vectors.utm_crs(vect) as utm_vect:
        >>>     utm_centroid = utm_vect.centroid
        >>>     utm_vect["centroid_utm"] = utm_centroid
        >>> vect["centroid_utm"].equals(c2)
        True
    """
    src_crs = None
    if not gdf.crs.is_projected:
        src_crs = gdf.crs
        gdf.to_crs(gdf.estimate_utm_crs(), inplace=True)
    try:
        yield gdf
    finally:
        if src_crs is not None:
            gdf.to_crs(src_crs, inplace=True)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the files"""

import os
import shutil
from datetime import date, datetime

import numpy as np
import pytest
from lxml import etree, html

from ci.script_utils import Polarization, files_path, s3_env
from sertit import AnyPath, ci, files, path, vectors

ci.reduce_verbosity()


@s3_env
def test_archive(tmp_path):
    """Test extracting functions"""
    # Archives
    zip_file = files_path().joinpath("test_zip.zip")
    zip2_file = files_path().joinpath("test_zip.zip")  # For overwrite
    zip_without_directory = files_path().joinpath("test_zip_without_directory.zip")
    tar_file = files_path().joinpath("test_tar.tar")
    tar_gz_file = files_path().joinpath("test_targz.tar.gz")

    # Core dir
    core_dir = files_path().joinpath("core")
    folder = core_dir
    archives = [
        zip_file,
        tar_file,
        tar_gz_file,
        folder,
        zip2_file,
        zip_without_directory,
    ]

    # Extract
    extracted_dirs = files.extract_files(archives, tmp_path, overwrite=True)
    files.extract_files([zip2_file], tmp_path, overwrite=False)  # Already existing

    # Test
    for ex_dir in extracted_dirs:
        ci.assert_dir_equal(core_dir, ex_dir)

    # Archive
    archive_base = os.path.join(tmp_path, "archive")
    for fmt in ["zip", "tar", "gztar"]:
        archive_fn = files.archive(
            folder_path=core_dir, archive_path=archive_base, fmt=fmt
        )
        out = files.extract_file(archive_fn, tmp_path)
        # an additional folder is created
        out_dir = path.listdir_abspath(out)[0]
        ci.assert_dir_equal(core_dir, out_dir)

        # Remove out directory in order to avoid any interferences
        files.remove(out)

    # Add to zip
    zip_out = zip2_file if path.is_cloud_path(zip2_file) else archive_base + ".zip"
    core_copy = files.copy(core_dir, os.path.join(tmp_path, "core2"))
    zip_out = files.add_to_zip(zip_out, core_copy)

    # Non existing
    with pytest.raises(FileNotFoundError):
        files.add_to_zip("not_existing.zip", core_copy)

    # Extract
    unzip_out = os.path.join(tmp_path, "out")
    unzip_out = files.extract_file(zip_out, unzip_out)

    # Test
    unzip_dirs = path.listdir_abspath(unzip_out)

    assert len(unzip_dirs) == 2
    ci.assert_dir_equal(unzip_dirs[0], unzip_dirs[1])


@s3_env
def test_archive_cloud(tmp_path):
    landsat_name = "LM05_L1TP_200030_20121230_20200820_02_T2_CI"
    folder = files_path().joinpath(landsat_name)
    arch_path = files.archive(folder, tmp_path / landsat_name)
    ci.assert_val(arch_path.name, landsat_name + ".zip", "archive name")


def test_archived_err(tmp_path):
    txt_path = tmp_path / "hello.txt"
    txt_path.write_text("test")

    with pytest.raises(TypeError):
        files.extract_file(txt_path, tmp_path)


@s3_env
def test_archived_files(tmp_path):
    landsat_name = "LM05_L1TP_200030_20121230_20200820_02_T2_CI"
    ok_folder = files_path().joinpath(landsat_name)
    zip_file = files_path().joinpath(f"{landsat_name}.zip")
    tar_file = files_path().joinpath(f"{landsat_name}.tar")
    targz_file = files_path().joinpath(f"{landsat_name}.tar.gz")
    sz_file = files_path().joinpath(f"{landsat_name}.7z")

    # VECTORS
    vect_name = "map-overlay.kml"
    vec_ok_path = ok_folder.joinpath(vect_name)
    if shutil.which("ogr2ogr"):  # Only works if ogr2ogr can be found.
        vect_regex = f".*{vect_name}"
        vect_zip = vectors.read(zip_file, archive_regex=vect_regex)
        vect_tar = vectors.read(tar_file, archive_regex=r".*overlay\.kml")
        vect_ok = vectors.read(vec_ok_path)
        assert not vect_ok.empty
        ci.assert_geom_equal(vect_ok, vect_zip)
        ci.assert_geom_equal(vect_ok, vect_tar)

    # XML
    xml_name = "LM05_L1TP_200030_20121230_20200820_02_T2_MTL.xml"
    xml_ok_path = ok_folder.joinpath(xml_name)
    if path.is_cloud_path(files_path()):
        xml_ok_path = str(xml_ok_path.download_to(tmp_path))
    else:
        xml_ok_path = str(xml_ok_path)

    xml_regex = f".*{xml_name}"
    xml_zip = files.read_archived_xml(zip_file, xml_regex)
    xml_tar = files.read_archived_xml(tar_file, r".*_MTL\.xml")
    xml_ok = etree.parse(xml_ok_path).getroot()
    ci.assert_xml_equal(xml_ok, xml_zip)
    ci.assert_xml_equal(xml_ok, xml_tar)

    # FILE + HTML
    html_zip_file = files_path().joinpath("productPreview.zip")
    html_tar_file = files_path().joinpath("productPreview.tar")
    html_name = "productPreview.html"
    html_ok_path = files_path().joinpath(html_name)
    if path.is_cloud_path(files_path()):
        html_ok_path = str(html_ok_path.download_to(tmp_path))
    else:
        html_ok_path = str(html_ok_path)

    html_regex = f".*{html_name}"

    # FILE
    file_zip = files.read_archived_file(html_zip_file, html_regex)
    file_tar = files.read_archived_file(html_tar_file, html_regex)
    html_ok = html.parse(html_ok_path).getroot()
    ci.assert_html_equal(html_ok, html.fromstring(file_zip))
    ci.assert_html_equal(html_ok, html.fromstring(file_tar))

    file_list = path.get_archived_file_list(html_zip_file)
    ci.assert_html_equal(
        html_ok,
        html.fromstring(
            files.read_archived_file(html_zip_file, html_regex, file_list=file_list)
        ),
    )

    # HTML
    html_zip = files.read_archived_html(html_zip_file, html_regex)
    html_tar = files.read_archived_html(html_tar_file, html_regex)
    ci.assert_html_equal(html_ok, html_zip)
    ci.assert_html_equal(html_ok, html_tar)
    ci.assert_html_equal(
        html_ok,
        files.read_archived_html(
            html_tar_file,
            html_regex,
            file_list=path.get_archived_file_list(html_tar_file),
        ),
    )

    # ERRORS
    with pytest.raises(TypeError):
        files.read_archived_file(targz_file, xml_regex)
    with pytest.raises(TypeError):
        files.read_archived_file(sz_file, xml_regex)
    with pytest.raises(FileNotFoundError):
        files.read_archived_file(zip_file, "cdzeferf")


def test_cp_rm(tmp_path):
    """Test CP/RM functions"""
    empty_tmp = os.listdir(tmp_path)

    # Copy file
    curr_path = os.path.realpath(__file__)
    file_1 = files.copy(curr_path, tmp_path)

    # Copy a folder onto a file: OSError: Copy error: File exists
    with pytest.raises(OSError):
        files.copy(tmp_path, file_1)

    file_2 = files.copy(curr_path, os.path.join(tmp_path, "test_pattern.py"))

    # Copy dir
    dir_path = os.path.dirname(curr_path)
    test_dir = files.copy(dir_path, os.path.join(tmp_path, os.path.basename(dir_path)))

    # Test copy
    assert os.path.isfile(file_1)
    assert os.path.isfile(file_2)
    assert os.path.isdir(test_dir)

    # Remove file
    files.remove(file_1)
    files.remove("non_existing_file.txt")
    files.remove_by_pattern(tmp_path, name_with_wildcard="*pattern*", extension="py")

    # Remove dir
    files.remove(test_dir)

    # Assert tempfile is empty
    assert os.listdir(tmp_path) == empty_tmp


@s3_env
def test_copy_cloud(tmp_path):
    zip_file = files_path().joinpath("test_zip.zip")
    zip_copy = files.copy(zip_file, tmp_path)
    assert zip_copy.is_file()
    assert zip_copy.name == "test_zip.zip"


def test_json(tmp_path):
    """Test json functions"""

    test_dict = {
        "A": 3,
        "C": "m2",  # Can be parsed as a date, we do not want that !
        "D": datetime.today(),
        "Dbis": date.today(),
        "E": np.int64(15),
        "F": Polarization.vv,
        "G": True,
        "H": AnyPath("/home/data"),
        "U": 3.1415,
    }

    json_file = os.path.join(tmp_path, "test.json")

    # Save JSON
    files.save_json(test_dict, json_file)

    # Load JSON
    obj = files.read_json(json_file)

    assert (
        obj.pop("F") == test_dict.pop("F").value
    )  # Enum are stored following their value

    assert obj.pop("H") == str(
        test_dict.pop("H")
    )  # Enum are stored following their value
    assert obj == test_dict


def test_pickle(tmp_path):
    """Test pickle functions"""

    test_dict = {
        "A": 3,
        "B": np.zeros((3, 3)),
        "C": "str",
        "D": datetime.today(),
        "E": np.int64(15),
    }

    pkl_file = os.path.join(tmp_path, "test.pkl")

    # Save pickle
    files.save_obj(test_dict, pkl_file)

    # Load pickle
    obj = files.load_obj(pkl_file)

    # Test (couldn't compare the dicts as they contain numpy arrays)
    np.testing.assert_equal(obj, test_dict)


def test_hash_file_content():
    """Test hash_file_content"""
    file_content = "This is a test."

    # Test
    hashed = files.hash_file_content(file_content)

    # Test
    assert hashed == "16c5bf1fc5"



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing SNAP functions"""

import shutil

import pytest

from sertit import ci, misc, snap

ci.reduce_verbosity()


def test_b2snap():
    """Testing SNAP functions"""
    ci.assert_val(snap.bytes2snap(32000), "31K", "32000")
    ci.assert_val(snap.bytes2snap(8), "8B", "8")


@pytest.mark.skipif(
    shutil.which("gpt") is None, reason="Only works if SNAP GPT's exe can be found."
)
def test_snap():
    # Do not test everything here, depends on the computer...
    cli = snap.get_gpt_cli("graph_path", display_snap_opt=True)
    tile_size = 512
    must_appear = [
        "gpt",
        '"graph_path"',
        "-q",
        "-J-Dsnap.log.level=WARNING",
        f"-J-Dsnap.jai.defaultTileSize={tile_size}",
        f"-J-Dsnap.dataio.reader.tileWidth={tile_size}",
        f"-J-Dsnap.dataio.reader.tileHeight={tile_size}",
        "-J-Dsnap.jai.prefetchTiles=true",
    ]

    for substr in must_appear:
        assert substr in cli


def test_snap_version():
    """Test SNAP version"""
    snap_version = snap.get_snap_version()
    assert misc.compare_version(snap_version, "13.0.0", "=="), (
        f"Unexpected SNAP version: {snap_version}."
    )



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Miscellaneous Tools"""

import logging
import os
import pprint
import subprocess
import sys
from contextlib import contextmanager
from enum import Enum, unique
from typing import Any

from packaging.version import Version

from sertit import AnyPath
from sertit.logs import SU_NAME
from sertit.types import AnyPathStrType

LOGGER = logging.getLogger(SU_NAME)


@unique
class ListEnum(Enum):
    """
    List Enum (enum with function listing names and values)

    Example:
        >>> @unique
        >>> class TsxPolarization(ListEnum):
        >>>     SINGLE = "S"  # Single
        >>>     DUAL = "D"  # Dual
        >>>     QUAD = "Q"  # Quad
        >>>     TWIN = "T"  # Twin
    """

    @classmethod
    def list_values(cls) -> list:
        """
        Get the value list of this enum

        Example:
            >>> TsxPolarization.list_values()
            ["S", "D", "Q", "T"]

        """
        return list(map(lambda c: c.value, cls))

    @classmethod
    def list_names(cls) -> list:
        """
        Get the name list of this enum:

        Example:
            >>> TsxPolarization.list_values()
            ["SINGLE", "DUAL", "QUAD", "TWIN"]
        """
        return list(map(lambda c: c.name, cls))

    @classmethod
    def from_value(cls, val: Any) -> "ListEnum":
        """
        Get the enum class from its value:

        Args:
            val (Any): Value of the Enum

        Returns:
            ListEnum: Enum with value

        Example:
            >>> TsxPolarization.from_value("Q")
            <TsxPolarization.QUAD: 'Q'>
        """
        if isinstance(val, cls):
            val = val.value
        try:
            return next(enum for enum in cls if enum.value == val)
        except StopIteration as ex:
            raise ValueError(f"Non existing {val} in {cls.list_values()}") from ex

    @classmethod
    def convert_from(cls, to_convert: list | str) -> list:
        """
        Convert from a list or a string to an enum instance

        Args:
            to_convert (list | str): List or string to convert into an enum instance

        Returns:
            list: Converted list

        Example:
            >>> TsxPolarization.convert_from(["SINGLE", "S", TsxPolarization.QUAD])
            [<TsxPolarization.SINGLE: 'S'>, <TsxPolarization.SINGLE: 'S'>, <TsxPolarization.QUAD: 'Q'>]

        """
        if not isinstance(to_convert, list):
            to_convert = [to_convert]

        enums = []

        for tc in to_convert:
            if tc in cls.list_values():
                enums.append(cls.from_value(tc))
            elif tc in cls.list_names():
                enums.append(getattr(cls, tc))
            elif isinstance(tc, cls):
                enums.append(tc)
            else:
                raise TypeError(
                    f"Invalid name {tc}, "
                    f"should be chosen among {cls.list_values()} or {cls.list_names()}"
                )

        return enums


def unique(sequence: list):
    """
    Keep only unique values from a list (any Iterable should work).

    Preserves the order of the sequence (except for sets of course).

    Args:
        sequence (list): List from which to keep only the unique values

    Returns:
        list: List containing only unique values

    Examples:
        >>> # With a list
        >>> unique([5, 4, 1, 2, 3, 1, 2])
        [5, 4, 1, 2, 3]

        >>> # With an array
        >>> unique(np.array([5, 4, 1, 2, 3, 1, 2]))
        [5, 4, 1, 2, 3]

        >>> # With a set (sorts the values, as a set would do!)
        >>> unique({5, 4, 1, 2, 3, 1, 2})
        [1, 2, 3, 4, 5]


    """
    return list(dict.fromkeys(sequence))


def remove_empty_values(
    object_with_empty_values: list | dict, other_empty_values: list = None
) -> list | dict:
    """
    Remove empty values from list.

    Args:
        object_with_empty_values (list): List with empty values

    Returns:
        list: Curated list

    Example:
        >>> lst = ["A", "T", "R", "", 3, None]
        >>> list_to_dict(lst)
        ["A", "T", "R", 3]

    """
    if not other_empty_values:
        other_empty_values = []

    if isinstance(object_with_empty_values, dict):
        return {
            k: v
            for k, v in object_with_empty_values.items()
            if v is not None and v not in other_empty_values
        }
    else:
        return [
            v
            for v in filter(None, object_with_empty_values)
            if v not in other_empty_values
        ]


def select_dict(dict_to_select: dict, keys_to_select: list) -> dict:
    """
    Select keys from a dictionary.

    Args:
        dict_to_select (dict): Dictionary to select from
        keys_to_select (list): List of keys to select in the dictionary

    Returns:
        dict: Dictionary with selected keys
    """
    return {k: v for k, v in dict_to_select.items() if k in keys_to_select}


def prune_dict(dict_to_prune: dict, keys_to_prune: list) -> dict:
    """
    Prune keys from a dictionary.

    Args:
        dict_to_prune (dict): Dictionary to prune from
        keys_to_prune (list): List of keys to prune in the dictionary

    Returns:
        dict: Dictionary with pruned keys
    """
    return {k: v for k, v in dict_to_prune.items() if k not in keys_to_prune}


def list_to_dict(dict_list: list) -> dict:
    """
    Return a dictionary from a list :code:`[key, value, key_2, value_2...]`

    Args:
        dict_list (list[str]): Dictionary as a list

    Returns:
        dict: Dictionary

    Example:
        >>> lst = ["A","T", "R", 3]
        >>> list_to_dict(lst)
        {"A": "T", "R": 3}
    """
    dictionary = {dict_list[i]: dict_list[i + 1] for i in range(0, len(dict_list), 2)}
    return dictionary


def nested_set(dic: dict, keys: list, value: Any) -> None:
    """
    Set value in nested directory:

    Args:
        dic (dict): Dictionary
        keys (list[str]): Keys as a list
        value: Value to be set

    Example:
        >>> dct = {"A": "T", "R": 3}
        >>> nested_set(dct, keys=["B", "C", "D"], value="value")
        {
        "A": "T",
        "R": 3,
        "B": {
             "C": {
                  "D": "value"
                  }
             }
        }
    """
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value


def check_mandatory_keys(data_dict: dict, mandatory_keys: list) -> None:
    """
    Check all mandatory argument in a dictionary.
    Raise an exception if a mandatory argument is missing.

    **Note**: nested keys do not work here !


    Args:
        data_dict (dict): Data dictionary to be checked
        mandatory_keys (list[str]): List of mandatory keys

    Example:
        >>> dct = {"A": "T", "R": 3}
        >>> check_mandatory_keys(dct, ["A", "R"])  # Returns nothing, is OK
        >>> check_mandatory_keys(dct, ["C"])
        Traceback (most recent call last):
          File "<input>", line 1, in <module>
          File "<input>", line 167, in check_mandatory_keys
        ValueError: Missing mandatory key 'C' among {'A': 'T', 'R': 3}
    """

    for mandatory_key in mandatory_keys:
        if mandatory_key not in data_dict:
            raise ValueError(
                f"Missing mandatory key '{mandatory_key}' among {pprint.pformat(data_dict)}"
            )


def find_by_key(data: dict, target: str) -> Any:
    """
    Find a value by key in a dictionary.

    Args:
        data (dict): Dict to walk through
        target (str): target key

    Returns:
        Any: Value data[...][target]

    Example:
        >>> dct = {
        >>>         "A": "T",
        >>>         "R": 3,
        >>>         "B": {
        >>>             "C": {
        >>>                 "D": "value"
        >>>                  }
        >>>              }
        >>>        }
        >>> find_by_key(dct, "D")
        "value"

    """
    val = None
    for key, value in data.items():
        if isinstance(value, dict):
            val = find_by_key(value, target)
            if val:
                break
        elif key == target:
            val = value
    return val


def run_cli(
    cmd: list | str,
    timeout: float = None,
    check_return_value: bool = True,
    in_background: bool = True,
    cwd="/",
) -> (int, str):
    """
    Run a command line.

    Args:
        cmd (str or list[str]): Command as a list
        timeout (float): Timeout
        check_return_value (bool): Check output value of the exe
        in_background (bool): Run the subprocess in background
        cwd (str): Working directory

    Returns:
        int, str: return value and output log

    Example:
        >>> cmd_hillshade = ["gdaldem", "--config",
        >>>                  "NUM_THREADS", "1",
        >>>                  "hillshade", strings.to_cmd_string(dem_path),
        >>>                  "-compute_edges",
        >>>                  "-z", self.nof_threads,
        >>>                  "-az", azimuth,
        >>>                  "-alt", zenith,
        >>>                  "-of", "GTiff",
        >>>                  strings.to_cmd_string(hillshade_dem)]
        >>> # Run command
        >>> run_cli(cmd_hillshade)
    """
    if isinstance(cmd, list):
        cmd = [str(cmd_i) for cmd_i in cmd]
        cmd_line = " ".join(cmd)
    elif isinstance(cmd, str):
        cmd_line = cmd
    else:
        raise TypeError("The command line should be given as a str or a list")

    # Background
    LOGGER.debug(cmd_line)
    if in_background:
        stdout = None
        stderr = None
        close_fds = True
    else:
        stdout = subprocess.PIPE
        stderr = subprocess.STDOUT
        close_fds = False

    # The os.setsid() is passed in the argument preexec_fn so
    # it's run after the fork() and before  exec() to run the shell.
    with subprocess.Popen(
        cmd_line,
        shell=True,
        stdout=stdout,
        stderr=stderr,
        cwd=cwd,
        start_new_session=True,
        close_fds=close_fds,
    ) as process:
        output = ""
        if not in_background:
            for line in process.stdout:  # pragma: no cover
                line = line.decode(
                    encoding=sys.stdout.encoding,
                    errors=(
                        "replace" if sys.version_info < (3, 5) else "backslashreplace"
                    ),
                ).rstrip()
                LOGGER.info(line)
                output += line

        # Get return value
        retval = process.wait(timeout)

        # Kill process
        process.kill()

    # Check return value
    if check_return_value and retval != 0:
        raise RuntimeError(f"Exe {cmd[0]} has failed.")  # pragma: no cover

    return retval, output


def get_function_name() -> str:
    """
    Get the name of the function where this one is launched.

    Returns:
        str: Function's name

    Example:
        >>> def huhuhu():
        >>>     return get_function_name()
        >>> huhuhu()
        "huhuhu"
    """
    # pylint: disable=W0212
    return sys._getframe(1).f_code.co_name


def in_docker() -> bool:
    """
    Check if the session is running inside a docker

    Returns:
        bool: True if inside a docker

    Example:
        >>> if in_docker():
        >>>    print("OMG we are stock in a Docker ! Get me out of here !")
        >>> else:
        >>>    print("We are safe")
    """
    try:
        with open("/proc/1/cgroup") as ifh:
            in_dck = "docker" in ifh.read()
    except Exception:  # pragma: no cover
        in_dck = False

    return in_dck


@contextmanager
def chdir(newdir: AnyPathStrType) -> None:
    """
    Change current directory, used as a context manager, i.e.:

    Args:
        newdir (str): New directory

    Example:
        >>> folder = r"C:/"
        >>> with chdir(folder):
        >>>     print(os.getcwd())
        'C:/'
    """
    newdir = AnyPath(newdir)
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def compare(a, b, operation: str) -> bool:
    """
    Compare two objects using a specific operation.
    Using this function allows to ask the user the operation he wants (see compare_version for example)

    Args:
        a: First object
        b: Second object
        operator (str): Operator to use (:code:`>`, :code:`<`, :code:`>=`, :code:`<=`, :code:`==`)

    Returns:
        bool: True if the comparison between the two objects is respected

    Example:
        >>> compare(1, 2, ">=")
        False

    """
    import operator

    ops = {
        ">": operator.gt,
        "<": operator.lt,
        ">=": operator.ge,
        "<=": operator.le,
        "==": operator.eq,
    }
    return ops[operation](a, b)


def compare_version(lib: str | Version, version_to_check: str, operator: str) -> bool:
    """
    Compare the version of a librarie to a reference, giving the operator.

    Args:
        lib (str | Version): Name of the library, it's version as a string or as a Version object
        version_to_check (str): Version of the library to be compared
        operator (str): Operator to use (:code:`>`, :code:`<`, :code:`>=`, :code:`<=`, :code:`==`)

    Returns:
        bool: True if the comparison between the version of the library and the reference version is respected

    Example:
        >>> compare_version("geopandas", "0.10.0", ">=")
        True
        >>> compare_version(sertit.__version__, "1.0.0", ">=")
        True

    """
    from importlib.metadata import PackageNotFoundError, version

    if isinstance(lib, Version):
        lib_version = lib
    elif isinstance(lib, str):
        try:
            lib_version = Version(version(lib))
        except PackageNotFoundError:
            lib_version = Version(lib)
    else:
        raise TypeError(
            "'lib' should either be the name of your library as a string or directly a 'Version' object."
        )

    return compare(lib_version, Version(version_to_check), operator)



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
CI tools
"""

import contextlib
import filecmp
import logging
import pprint
from doctest import Example
from typing import Any

import geopandas as gpd
import numpy as np
from lxml import etree, html
from lxml.doctestcompare import LHTMLOutputChecker, LXMLOutputChecker
from shapely import force_2d, normalize
from shapely.testing import assert_geometries_equal

from sertit import AnyPath, dask, files, rasters, s3
from sertit.logs import SU_NAME
from sertit.types import AnyPathStrType, AnyVectorType, AnyXrDataStructure

LOGGER = logging.getLogger(SU_NAME)

# Alias for compatibility (don't deprecate them)
AWS_ACCESS_KEY_ID = s3.AWS_ACCESS_KEY_ID
AWS_SECRET_ACCESS_KEY = s3.AWS_SECRET_ACCESS_KEY
AWS_S3_ENDPOINT = s3.AWS_S3_ENDPOINT


def assert_val(val_1: Any, val_2: Any, field: str) -> None:
    """
    Compare two values corresponding to a field

    Args:
        val_1 (Any): Value 1
        val_2 (Any): Value 2
        field (str): Field to compare
    """
    desc = f"{field} incoherent:\n{val_1} != {val_2}"

    test_done = False

    # Manage None as value
    if val_2 is None or val_1 is None:
        assert val_1 is val_2, desc
        test_done = True
    else:
        with contextlib.suppress(TypeError):
            if np.isnan(val_1).all() or np.isnan(val_2).all():
                assert np.isnan(val_1).all() and np.isnan(val_2).all(), desc
                test_done = True

        if not test_done:
            with contextlib.suppress(ValueError):
                assert val_1 == val_2, desc
                test_done = True

        if not test_done:
            with contextlib.suppress(ValueError):
                assert all(val_1 == val_2), desc
                test_done = True

        if not test_done:  # pragma: no cover
            # xarray.DataArrayCoordinates are failing
            assert val_1.equals(val_2), desc
            test_done = True

    if not test_done:  # pragma: no cover
        raise AssertionError(
            f"{desc} - Test hasn't been done. Please write an issue about that."
        )


def assert_field(dict_1: dict, dict_2: dict, field: str) -> None:
    """
    Compare two fields of a dictionary

    Args:
        dict_1 (dict): Dict 1
        dict_2 (dict): Dict 2
        field (str): Field to compare
    """
    assert_val(dict_1[field], dict_2[field], field)


def assert_files_equal(file_1: AnyPathStrType, file_2: AnyPathStrType):
    """
    Assert to files are equal by hashing its content

    Args:
        file_1 (str): Path to file 1
        file_2 (str): Path to file 2
    """
    with AnyPath(file_1).open("r") as f1, AnyPath(file_2).open("r") as f2:
        assert files.hash_file_content(f1.read()) == files.hash_file_content(f2.read())


def assert_meta(meta_1: dict, meta_2: dict, tf_precision: float = 1e-9):
    """
    Compare rasterio metadata

    Args:
        meta_1 (dict): Metadata 1
        meta_2 (dict): Metadata 2
        tf_precision (float): Transform precision (in transform units)
    """
    # Driver
    assert_field(meta_1, meta_2, "driver")
    assert_field(meta_1, meta_2, "dtype")
    assert_field(meta_1, meta_2, "nodata")
    assert_field(meta_1, meta_2, "width")
    assert_field(meta_1, meta_2, "height")
    assert_field(meta_1, meta_2, "count")
    assert_field(meta_1, meta_2, "crs")

    assert meta_1["transform"].almost_equals(
        meta_2["transform"], precision=tf_precision
    ), f"transform incoherent:\n{meta_1['transform']}\n!=\n{meta_2['transform']}"


def assert_raster_equal(path_1: AnyPathStrType, path_2: AnyPathStrType) -> None:
    """
    Assert that two rasters are equal.

    Useful for pytests.

    Args:
        path_1 (AnyPathStrType): Raster 1
        path_2 (AnyPathStrType): Raster 2

    Example:
        >>> path = r"CI/DATA/rasters/raster.tif"
        >>> assert_raster_equal(path, path)
        >>> # Raises AssertionError if sth goes wrong
    """
    try:
        import rasterio
    except ModuleNotFoundError as ex:  # pragma: no cover
        raise ModuleNotFoundError(
            "Please install 'rasterio' to use assert_raster_equal."
        ) from ex

    with rasterio.open(str(path_1)) as ds_1, rasterio.open(str(path_2)) as ds_2:
        # Metadata
        assert_meta(ds_1.meta, ds_2.meta)

        # Assert equal
        np.testing.assert_array_equal(ds_1.read(), ds_2.read())


def assert_raster_almost_equal(
    path_1: AnyPathStrType, path_2: AnyPathStrType, decimal=7
) -> None:
    """
    Assert that two rasters are almost equal.
    (everything is equal except the transform and the arrays that are almost equal)

    Accepts an offset of :code:`1E{decimal}` on the array and a precision of 10^-{decimal} on the transform

    Useful for pytests.

    Args:
        path_1 (AnyPathStrType): Raster 1
        path_2 (AnyPathStrType): Raster 2
        decimal (int): Number of decimal

    Example:
        >>> path = r"CI/DATA/rasters/raster.tif"
        >>> path2 = r"CI/DATA/rasters/raster_almost.tif"
        >>> assert_raster_almost_equal(path, path2)
        >>> # Raises AssertionError if sth goes wrong
    """
    try:
        import rasterio
    except ModuleNotFoundError as ex:  # pragma: no cover
        raise ModuleNotFoundError(
            "Please install 'rasterio' to use assert_raster_almost_equal."
        ) from ex

    with rasterio.open(str(path_1)) as ds_1, rasterio.open(str(path_2)) as ds_2:
        # Metadata
        assert_meta(ds_1.meta, ds_2.meta, tf_precision=10**-decimal)

        # Assert almost equal
        errors = []
        for i in range(ds_1.count):
            desc = (
                f": {ds_1.descriptions[i]}" if ds_1.descriptions[i] is not None else ""
            )
            LOGGER.info(f"Checking Band {i + 1}{desc}")
            try:
                marr_1 = ds_1.read(i + 1, masked=True)
                marr_2 = ds_2.read(i + 1, masked=True)

                # Set nodata
                marr_1 = np.where(marr_1.mask, np.nan, marr_1)
                marr_2 = np.where(marr_2.mask, np.nan, marr_2)
                np.testing.assert_array_almost_equal(marr_1, marr_2, decimal=decimal)
            except AssertionError:
                text = f"Band {i + 1}{desc} failed"
                errors.append(text)
                LOGGER.error(text, exc_info=True)

        if errors:
            raise AssertionError(errors)


def assert_raster_almost_equal_magnitude(
    path_1: AnyPathStrType, path_2: AnyPathStrType, decimal=2
) -> None:
    """
    Assert that two rasters are almost equal, with the decimal taken on the scientif representation of the array.
    (everything is equal except the transform and the arrays that are almost equal)

    Accepts an offset of :code:`1E{decimal}` on the array divided by the order of magnitude of the array and a precision of 10^-{decimal} on the transform

    i.e. `decimal=2, mean(array) = 15.2, true = 13.2687, false = 13.977, comparison: 1.32687, false <-> 1.3977 => false (2 != 9)`

    Useful for pytests.

    Args:
        path_1 (AnyPathStrType): Raster 1
        path_2 (AnyPathStrType): Raster 2
        decimal (int): Number of decimal

    Example:
        >>> path = r"CI/DATA/rasters/raster.tif"
        >>> path2 = r"CI/DATA/rasters/raster_almost.tif"
        >>> assert_raster_almost_equal_magnitude(path, path2)
        >>> # Raises AssertionError if sth goes wrong
    """
    try:
        import rasterio
    except ModuleNotFoundError as ex:  # pragma: no cover
        raise ModuleNotFoundError(
            "Please install 'rasterio' to use assert_raster_almost_equal."
        ) from ex

    with rasterio.open(str(path_1)) as ds_1, rasterio.open(str(path_2)) as ds_2:
        # Metadata
        assert_meta(ds_1.meta, ds_2.meta, tf_precision=10**-decimal)

        # Assert almost equal
        errors = []
        for i in range(ds_1.count):
            desc = (
                f": {ds_1.descriptions[i]}" if ds_1.descriptions[i] is not None else ""
            )
            LOGGER.info(f"Checking Band {i + 1}{desc}")
            try:
                marr_1 = ds_1.read(i + 1, masked=True)
                marr_2 = ds_2.read(i + 1, masked=True)

                # Set nodata
                marr_1 = np.where(marr_1.mask, np.nan, marr_1)
                marr_2 = np.where(marr_2.mask, np.nan, marr_2)

                # Check if arrays are identical (including NaN positions)
                if np.array_equal(marr_1, marr_2, equal_nan=True):
                    continue

                # Get non-NaN values for magnitude calculation
                valid_mask_1 = ~np.isnan(marr_1)
                valid_mask_2 = ~np.isnan(marr_2)

                # Check if both arrays have the same NaN pattern
                if not np.array_equal(valid_mask_1, valid_mask_2):
                    raise AssertionError("Arrays have different NaN patterns")

                # If all values are NaN, arrays should be identical
                if not np.any(valid_mask_1):
                    if not np.array_equal(marr_1, marr_2, equal_nan=True):
                        raise AssertionError("All-NaN arrays are not identical")
                    continue

                # Calculate magnitude using valid (non-NaN) values only
                valid_values_1 = marr_1[valid_mask_1]
                valid_values_2 = marr_2[valid_mask_2]

                # Use the median of valid values for magnitude calculation
                median_val = np.median(np.abs(valid_values_1))

                if median_val == 0:
                    magnitude = 0
                else:
                    magnitude = np.floor(np.log10(median_val))
                    if np.isinf(magnitude) or np.isnan(magnitude):
                        magnitude = 0

                # Create arrays for comparison (preserve NaN values)
                comparison_arr_1 = np.full_like(marr_1, np.nan)
                comparison_arr_2 = np.full_like(marr_2, np.nan)

                # Only compare valid values with magnitude adjustment
                comparison_arr_1[valid_mask_1] = valid_values_1 / (10**magnitude)
                comparison_arr_2[valid_mask_2] = valid_values_2 / (10**magnitude)

                np.testing.assert_array_almost_equal(
                    comparison_arr_1,
                    comparison_arr_2,
                    decimal=decimal,
                )

            except AssertionError:
                text = f"Band {i + 1}{desc} failed"
                errors.append(text)
                LOGGER.error(text, exc_info=True)

        if errors:
            raise AssertionError(errors)


def assert_raster_max_mismatch(
    path_1: AnyPathStrType,
    path_2: AnyPathStrType,
    max_mismatch_pct=0.5,
    decimal=-1,
) -> None:
    """
    Assert that two rasters are almost equal.
    (everything is equal except the transform and the arrays that are almost equal)

    Accepts an offset of :code:`1E{decimal}` on the array and a precision of 10^-9 on the transform

    Useful for pytests.

    Args:
        path_1 (AnyPathStrType): Raster 1
        path_2 (AnyPathStrType): Raster 2
        max_mismatch_pct (float): Maximum of element mismatch in %

    Example:
        >>> path = r"CI/DATA/rasters/raster.tif"
        >>> path2 = r"CI/DATA/rasters/raster_almost.tif"
        >>> assert_raster_max_mismatch(path, path2)
        >>> # Raises AssertionError if sth goes wrong
    """
    try:
        import rasterio
    except ModuleNotFoundError as ex:  # pragma: no cover
        raise ModuleNotFoundError(
            "Please install 'rasterio' to use assert_raster_max_mismatch."
        ) from ex

    with rasterio.open(str(path_1)) as ds_1, rasterio.open(str(path_2)) as ds_2:
        # Metadata
        assert_meta(ds_1.meta, ds_2.meta)

        # Compute the number of mismatch
        arr_1 = ds_1.read(masked=True)
        arr_2 = ds_2.read(masked=True)

        # Set nodata (to 0 as we will make a diff)
        arr_1 = np.where(arr_1.mask, 0, arr_1)
        arr_2 = np.where(arr_2.mask, 0, arr_2)

        if decimal >= 0:
            arr_1 = np.round(arr_1, decimal)
            arr_2 = np.round(arr_2, decimal)

        diffs = np.abs(arr_1 - arr_2)
        nof_mismatch = np.count_nonzero(diffs)
        nof_elements = ds_1.count * ds_1.width * ds_1.height
        pct_mismatch = nof_mismatch / nof_elements * 100.0
        assert pct_mismatch < max_mismatch_pct, (
            f"Too many mismatches !\n"
            f"Number of mismatches: {nof_mismatch} / {nof_elements},\n"
            f"Percentage of mismatches: {pct_mismatch:0.2f}% > {max_mismatch_pct}%\n"
            f"Mean of mismatches: {np.nanmean(np.where(diffs != 0, diffs, np.nan)):0.2f}"
        )


def assert_dir_equal(path_1: AnyPathStrType, path_2: AnyPathStrType) -> None:
    """
    Assert that two directories are equal.

    Useful for pytests.

    Args:
        path_1 (str): Directory 1
        path_2 (str): Directory 2

    Example:
        >>> path = r"CI/DATA/rasters"
        >>> assert_dir_equal(path, path)
        >>> # Raises AssertionError if sth goes wrong
    """
    path_1 = AnyPath(path_1)
    path_2 = AnyPath(path_2)
    assert path_1.is_dir(), f"{path_1} is not a directory!"
    assert path_2.is_dir(), f"{path_2} is not a directory!"

    dcmp = filecmp.dircmp(path_1, path_2)
    try:
        assert dcmp.left_only == [], (
            f"More files in {path_1}!\n{pprint.pformat(list(dcmp.left_only))}"
        )
        assert dcmp.right_only == [], (
            f"More files in {path_2}!\n{pprint.pformat(list(dcmp.right_only))}"
        )
    except FileNotFoundError:
        files_1 = [AnyPath(p).name for p in AnyPath(path_1).iterdir()]
        files_2 = [AnyPath(p).name for p in AnyPath(path_2).iterdir()]

        for f1 in files_1:
            assert f1 in files_2, (
                f"File missing!\n{f1} not in {pprint.pformat(files_2)}"
            )

        for f2 in files_2:
            assert f2 in files_1, (
                f"File missing!\n{f2} not in {pprint.pformat(files_1)}"
            )


def assert_geom_equal(
    geom_1: AnyVectorType,
    geom_2: AnyVectorType,
    ignore_z=True,
) -> None:
    """
    Assert that two geometries are equal
    (do not check equality between geodataframe as they may differ on other fields).

    Useful for pytests.

    Args:
        geom_1 (AnyVectorType): Geometry 1
        geom_2 (AnyVectorType): Geometry 2
        ignore_z (bool): Ignore Z coordinate

    Warning:
        Only checks:
         - valid geometries
         - length of GeoDataFrame
         - CRS

    Example:
        >>> path = r"CI/DATA/vectors/aoi.geojson"
        >>> assert_geom_equal(path, path)
        >>> # Raises AssertionError if sth goes wrong
    """
    try:
        from sertit import vectors
    except ModuleNotFoundError as ex:  # pragma: no cover
        raise ModuleNotFoundError(
            "Please install 'rasterio' and 'geopandas' to use the 'vectors' package."
        ) from ex

    if not isinstance(geom_1, (gpd.GeoDataFrame, gpd.GeoSeries)):
        geom_1 = vectors.read(geom_1)
    if not isinstance(geom_2, (gpd.GeoDataFrame, gpd.GeoSeries)):
        geom_2 = vectors.read(geom_2)

    assert len(geom_1) == len(geom_2), (
        f"Non equal geometry lengths!\n{len(geom_1)} != {len(geom_2)}"
    )
    assert geom_1.crs == geom_2.crs, (
        f"Non equal geometry CRS!\n{geom_1.crs} != {geom_2.crs}"
    )

    for idx in range(len(geom_1)):
        curr_geom_1 = normalize(geom_1.geometry.iat[idx])
        curr_geom_2 = normalize(geom_2.geometry.iat[idx])

        if ignore_z:
            curr_geom_1 = force_2d(curr_geom_1)
            curr_geom_2 = force_2d(curr_geom_2)

        # If valid geometries, assert that the both are equal
        if curr_geom_1.is_valid and curr_geom_2.is_valid:
            try:
                assert curr_geom_1.equals(curr_geom_2), (
                    f"Non equal geometries!\n{curr_geom_1} != {curr_geom_2}"
                )
            except AssertionError:
                # Get tolerance
                tol = 1e-7 if geom_1.crs.is_geographic else 1e-3

                # This functions tests differently than equals
                assert_geometries_equal(
                    curr_geom_1,
                    curr_geom_2,
                    tolerance=tol,
                    equal_none=False,
                    equal_nan=False,
                    normalize=True,
                )


def assert_geom_almost_equal(
    geom_1: AnyVectorType,
    geom_2: AnyVectorType,
    decimal=9,
    ignore_z=True,
) -> None:
    """
    Assert that two geometries are equal
    (do not check equality between geodataframe as they may differ on other fields).

    Useful for pytests.

    Args:
        geom_1 (AnyVectorType): Geometry 1
        geom_2 (AnyVectorType): Geometry 2
        decimal (int): Number of decimal
        ignore_z (bool): Ignore Z coordinate

    Warning:
        Only checks:
         - valid geometries
         - length of GeoDataFrame
         - CRS

    Example:
        >>> path = r"CI/DATA/vectors/aoi.geojson"
        >>> assert_geom_almost_equal(path, path)
        >>> # Raises AssertionError if sth goes wrong
    """
    try:
        from sertit import vectors
    except ModuleNotFoundError as ex:  # pragma: no cover
        raise ModuleNotFoundError(
            "Please install 'rasterio' and 'geopandas' to use the 'vectors' package."
        ) from ex

    if not isinstance(geom_1, (gpd.GeoDataFrame, gpd.GeoSeries)):
        geom_1 = vectors.read(geom_1)
    if not isinstance(geom_2, (gpd.GeoDataFrame, gpd.GeoSeries)):
        geom_2 = vectors.read(geom_2)

    assert len(geom_1) == len(geom_2), (
        f"Non equal geometry lengths!\n{len(geom_1)} != {len(geom_2)}"
    )
    assert geom_1.crs == geom_2.crs, (
        f"Non equal geometry CRS!\n{geom_1.crs} != {geom_2.crs}"
    )

    for idx in range(len(geom_1)):
        curr_geom_1 = normalize(geom_1.geometry.iat[idx])
        curr_geom_2 = normalize(geom_2.geometry.iat[idx])

        if ignore_z:
            curr_geom_1 = force_2d(curr_geom_1)
            curr_geom_2 = force_2d(curr_geom_2)

        # If valid geometries, assert that the both are equal
        if curr_geom_1.is_valid and curr_geom_2.is_valid:
            try:
                assert curr_geom_1.equals_exact(curr_geom_2, tolerance=10**-decimal), (
                    f"Non equal geometries!\n{curr_geom_1} != {curr_geom_2}"
                )
            except AssertionError:
                # This functions tests differently than equals
                assert_geometries_equal(
                    curr_geom_1,
                    curr_geom_2,
                    tolerance=10**-decimal,
                    equal_none=False,
                    equal_nan=False,
                    normalize=True,
                )


def assert_dim_file_equal(path_1: AnyPathStrType, path_2: AnyPathStrType) -> None:
    """
    Assert that two .dim files are equal.

    Useful for pytests.

    Args:
        path_1 (AnyPathStrType): Raster 1
        path_2 (AnyPathStrType): Raster 2


    Example:
        >>> path = r"CI/DATA/rasters/raster.dim"
        >>> assert_dim_file_equal(path, path)
        >>> # Raises AssertionError if sth goes wrong
    """
    # TODO : Add Checking if .dim
    list_img_path_1 = rasters.get_dim_img_path(path_1, get_list=True)
    list_img_path_2 = rasters.get_dim_img_path(path_2, get_list=True)

    assert len(list_img_path_1) == len(list_img_path_2), (
        f"Files don't have the same number of bands!\n{len(list_img_path_1)} != {len(list_img_path_2)}"
    )
    assert sorted([img_file.name for img_file in list_img_path_1]) == sorted(
        [img_file.name for img_file in list_img_path_2]
    ), (
        f"Bands are not the same!\n{sorted([img_file.name for img_file in list_img_path_1])} != {sorted([img_file.name for img_file in list_img_path_2])}"
    )

    if len(list_img_path_1) == 1:
        out_raster = list_img_path_1[0]
        expected_raster = list_img_path_2[0]
        assert_raster_equal(expected_raster, out_raster)

    else:
        list_img_path_1.sort()
        list_img_path_2.sort()
        for i in range(len(list_img_path_1)):
            out_raster = list_img_path_1[i]
            expected_raster = list_img_path_2[i]
            assert_raster_equal(expected_raster, out_raster)


def assert_xml_equal(xml_elem_1: etree._Element, xml_elem_2: etree._Element) -> None:
    """
    Assert that 2 XML (as etree Elements) are equal.

    -> Useful for pytests.

    Args:
        xml_elem_1 (etree._Element): 1st Element
        xml_elem_2 (etree._Element): 2nd Element
    """
    str_1 = etree.tostring(xml_elem_1, encoding="unicode")
    str_2 = etree.tostring(xml_elem_2, encoding="unicode")
    checker = LXMLOutputChecker()
    if not checker.check_output(str_1, str_2, 0):
        message = checker.output_difference(Example("", str_1), str_2, 0)
        raise AssertionError(message)


def assert_html_equal(xml_elem_1: etree._Element, xml_elem_2: etree._Element) -> None:
    """
    Assert that 2 XML (as etree Elements) are equal.

    -> Useful for pytests.

    Args:
        xml_elem_1 (etree._Element): 1st Element
        xml_elem_2 (etree._Element): 2nd Element
    """
    str_1 = html.tostring(xml_elem_1, encoding="unicode")
    str_2 = html.tostring(xml_elem_2, encoding="unicode")
    checker = LHTMLOutputChecker()
    if not checker.check_output(str_1, str_2, 0):
        message = checker.output_difference(Example("", str_1), str_2, 0)
        raise AssertionError(message)


def assert_xr_encoding_attrs(
    xda_1: AnyXrDataStructure,
    xda_2: AnyXrDataStructure,
    unchecked_attr: list | str = None,
):
    """
    Assert that the attributes and the encoding of xarray.DataArray/set are the same

    Args:
        xda_1 (AnyXrDataStructure): First xarray
        xda_2 (AnyXrDataStructure): Second xarray
        unchecked_attr (list | str): Don't check this list of attributes
    """
    if unchecked_attr is None:
        unchecked_attr = []
    elif not isinstance(unchecked_attr, list):
        unchecked_attr = [unchecked_attr]

    # Attributes
    try:
        LOGGER.debug(
            f"Checking {len(set(xda_1.attrs.keys()) | set(xda_2.attrs.keys()))} attributes."
        )
        assert xda_1.attrs == xda_2.attrs
    except AssertionError:
        try:
            for key, _ in xda_1.attrs.items():
                if not key.startswith("_") and key not in unchecked_attr:
                    assert xda_1.attrs[key] == xda_2.attrs[key], (
                        f"{xda_1.attrs[key]=} != {xda_2.attrs[key]=}"
                    )

            for key, _ in xda_2.attrs.items():
                if not key.startswith("_") and key not in unchecked_attr:
                    assert xda_1.attrs[key] == xda_2.attrs[key], (
                        f"{xda_1.attrs[key]=} != {xda_2.attrs[key]=}"
                    )
        except KeyError as exc:
            raise AssertionError(
                f"Missing '{exc}' in attributes of one DataArray/Dataset"
            ) from exc

    # Encoding
    try:
        LOGGER.debug(
            f"Checking {len(set(xda_1.encoding.keys()) | set(xda_2.encoding.keys()))} encoding attributes."
        )
        assert xda_1.encoding == xda_2.encoding
    except AssertionError:
        try:
            for key, _ in xda_1.encoding.items():
                if not key.startswith("_") and key not in unchecked_attr:
                    assert xda_1.encoding[key] == xda_2.encoding[key], (
                        f"{xda_1.encoding[key]=} != {xda_2.encoding[key]=}"
                    )

            for key, _ in xda_2.encoding.items():
                if not key.startswith("_") and key not in unchecked_attr:
                    assert xda_1.encoding[key] == xda_2.encoding[key], (
                        f"{xda_1.encoding[key]=} != {xda_2.encoding[key]=}"
                    )
        except KeyError as exc:
            raise AssertionError(
                f"Missing '{exc}' in encoding of one DataArray/Dataset"
            ) from exc


def reduce_verbosity(other_loggers: list = None) -> None:
    """
    Reduce verbosity for other loggers (setting them to WARNING)

        Args:
            other_loggers (list): Other loggers to reduce verosity

    """
    loggers = [
        "boto3",
        "botocore",
        "botocore.hooks",
        "botocore.auth",
        "shapely",
        "shapely.geos",
        "fiona",
        "rasterio",
        "urllib3",
        "s3transfer",
        "pyproj",
        "matplotlib",
        "asyncio",
        "bokeh",
        "numba",
        "distributed",
        "distributed.worker",
        "distributed.scheduler",
        "distributed.nanny",
        "distributed.core",
        "distributed.http.proxy",
        "distributed.batched",
    ]
    if other_loggers:
        loggers += other_loggers

    # Unique logger names
    for logger in list(set(loggers)):
        logging.getLogger(logger).setLevel(logging.WARNING)


def assert_chunked(result: AnyXrDataStructure) -> None:
    """
    Assert the result is still chunked.

    Args:
        result (AnyXrDataStructure): Result to check
    """
    assert dask.is_chunked(result), "Your data is not chunked!"


def assert_computed(result: AnyXrDataStructure) -> None:
    """
    Assert the result has been computed (i.e. the xarray structure is not chunked)

    Args:
        result (AnyXrDataStructure): Result to check
    """
    assert dask.is_computed(result), "Your data is not computed!"



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing vector functions"""

from ci.script_utils import KAPUT_KWARGS, geometry_path, s3_env, vectors_path
from sertit import ci, geometry, vectors
from sertit.geometry import (
    buffer,
    fill_polygon_holes,
    get_wider_exterior,
    intersects,
    line_merge,
    nearest_neighbors,
    split,
)

ci.reduce_verbosity()


@s3_env
def test_get_wider_exterior():
    """Test get_wider_exterior"""
    footprint_raw_path = geometry_path().joinpath("footprint_raw.geojson")
    footprint_path = geometry_path().joinpath("footprint.geojson")
    ci.assert_geom_equal(
        get_wider_exterior(vectors.read(footprint_raw_path)),
        vectors.read(footprint_path),
    )


@s3_env
def test_simplify_footprint():
    """Test simplify footprint"""
    complicated_footprint_path = geometry_path().joinpath(
        "complicated_footprint_spot6.geojson"
    )
    max_nof_vertices = 40
    complicated_footprint = vectors.read(complicated_footprint_path)
    ok_footprint = geometry.simplify_footprint(
        complicated_footprint, resolution=1.5, max_nof_vertices=max_nof_vertices
    )
    assert len(ok_footprint.geometry.exterior.iat[0].coords) < max_nof_vertices

    # Just to test
    nof_vertices_complicated = len(
        complicated_footprint.explode(index_parts=True).geometry.exterior.iat[0].coords
    )
    assert nof_vertices_complicated > max_nof_vertices


@s3_env
def test_geometry_fct():
    """Test other geometry functions"""
    kml_path = vectors_path().joinpath("aoi.kml")
    env = vectors.read(kml_path).envelope[0]
    from_env = geometry.from_bounds_to_polygon(*geometry.from_polygon_to_bounds(env))
    assert env.bounds == from_env.bounds


@s3_env
def test_make_valid():
    """Test make valid"""
    broken_geom_path = geometry_path().joinpath("broken_geom.shp")
    broken_geom = vectors.read(broken_geom_path)
    assert len(broken_geom[~broken_geom.is_valid]) == 1
    valid = geometry.make_valid(broken_geom, verbose=True)
    assert len(valid[~valid.is_valid]) == 0
    assert len(valid) == len(broken_geom)


@s3_env
def test_fill_polygon_holes():
    """Test fill_polygon_holes"""
    water_path = geometry_path().joinpath("water.geojson")
    water_none_path = geometry_path().joinpath("water_filled_none.geojson")
    water_0_path = geometry_path().joinpath("water_filled_0.geojson")
    water_1000_path = geometry_path().joinpath("water_filled_1000.geojson")
    water = vectors.read(water_path)

    ci.assert_geom_equal(fill_polygon_holes(water), vectors.read(water_none_path))
    ci.assert_geom_equal(fill_polygon_holes(water, 0), vectors.read(water_0_path))
    ci.assert_geom_equal(
        fill_polygon_holes(water, threshold=1000), vectors.read(water_1000_path)
    )


@s3_env
def test_split():
    """Test split"""
    water_path = geometry_path().joinpath("water.geojson")
    water = vectors.read(water_path)

    # No MultiLineStrings
    footprint_path = geometry_path().joinpath("footprint_split.geojson")
    water_split_path = geometry_path().joinpath("water_split.geojson")
    ci.assert_geom_equal(
        split(water, vectors.read(footprint_path)), vectors.read(water_split_path)
    )

    # With MultiLineStrings
    footprint_raw_path = geometry_path().joinpath("footprint_raw.geojson")
    water_split_raw_path = geometry_path().joinpath("water_split_raw.geojson")
    ci.assert_geom_equal(
        split(water, vectors.read(footprint_raw_path)),
        vectors.read(water_split_raw_path),
    )

    # Test with lines as splitter (with and without line_merge)
    # Without line_merge: doesn't split anything
    lines_raw_path = geometry_path().joinpath("lines.shp")
    lines_raw = vectors.read(lines_raw_path)
    ci.assert_geom_equal(
        split(water, lines_raw),
        water,
    )

    # With line_merge: works
    water_split_line_path = geometry_path().joinpath("water_split_line.geojson")
    lines_merged = line_merge(lines_raw)
    ci.assert_geom_equal(
        split(water, lines_merged),
        vectors.read(water_split_line_path),
    )


@s3_env
def test_intersects():
    """Test intersects"""
    water_path = geometry_path().joinpath("water.geojson")
    lakes_path = geometry_path().joinpath("lakes.geojson")

    inter = intersects(vectors.read(lakes_path), vectors.read(water_path))
    ci.assert_val(inter.index, [2, 3], "Index")

    inter = intersects(
        vectors.read(lakes_path), vectors.read(water_path), buffer_on_input=1000
    )
    ci.assert_val(inter.index, [0, 1, 2, 3], "Index")


@s3_env
def test_buffer():
    """Test buffer"""
    water_path = geometry_path().joinpath("water.geojson")
    water = vectors.read(water_path)
    buffer_true = water.copy()
    buffer_true.geometry = water.buffer(10)

    ci.assert_geom_equal(buffer(water, 10), buffer_true)


@s3_env
def test_nearest_neighbors():
    """Test nearest_neighbors"""
    src_path = geometry_path().joinpath("source.geojson")
    candidates_path = geometry_path().joinpath("candidates.geojson")

    src = vectors.read(src_path)
    candidates = vectors.read(candidates_path)

    # Radius
    radius = 100
    closest, distances = nearest_neighbors(
        src, candidates, method="radius", radius=radius, **KAPUT_KWARGS
    )
    for curr_closest, curr_dist in zip(closest, distances, strict=True):
        ci.assert_val(len(curr_closest), 1, "length")
        assert curr_dist[0] < radius, (
            f"distance superior to radius: {curr_dist[0]} > {radius}"
        )

    # Nof neighbors
    nof_neighbors = 1
    closest, distances = nearest_neighbors(
        src, candidates, method="k_neighbors", k_neighbors=nof_neighbors
    )
    for curr_closest, curr_dist in zip(closest, distances, strict=True):
        ci.assert_val(len(curr_closest), nof_neighbors, "length")
        assert curr_dist[0] < radius, (
            f"distance superior to wanted distance: {curr_dist[0]} > {radius}"
        )

    # Ensure it works with k > nof candidates
    closest, distances = nearest_neighbors(
        src, candidates, method="k_neighbors", k_neighbors=len(candidates) + 10
    )


def test_force_2_or_3d():
    """Force 2D or 3D"""
    aoi_kml = vectors.read(vectors_path().joinpath("aoi.kml"))

    assert aoi_kml.has_z.any()

    # Check equality
    ci.assert_geom_equal(
        aoi_kml, geometry.force_3d(geometry.force_2d(aoi_kml)), ignore_z=False
    )

    # Check Z type
    assert not geometry.force_2d(aoi_kml).has_z.any()
    assert geometry.force_3d(geometry.force_2d(aoi_kml)).has_z.any()



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the os_utils module"""

from sertit import ci, os_utils

ci.reduce_verbosity()


def test_str_containing_version():
    """Test str_containing_version function"""
    ci.assert_val(
        os_utils._compare_str_containing_version("qgis 3.34.5", "qgis 3.35.6"),
        -1,
        "Version <=",
    )
    ci.assert_val(
        os_utils._compare_str_containing_version("qgis 3.34.5", "qgis 3.34.5"),
        0,
        "Version ==",
    )
    ci.assert_val(
        os_utils._compare_str_containing_version("qgis 3.37.5", "qgis 3.35.6"),
        1,
        "Version >=",
    )


def test_qgis_bin():
    """Test qgis_bin function"""
    # We are on linux, this fct returns nothing
    ci.assert_val(os_utils.qgis_bin(), None, "Empty qgis_bin")


def test_gdalbuildvrt():
    """Test gdalbuildvrt function"""
    # We are on linux, this fct returns the name of the exe only
    ci.assert_val(os_utils.gdalbuildvrt_exe(), "gdalbuildvrt", "gdalbuildvrt on Linux")



# -*- coding: utf-8 -*-
# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing the arcpy module"""

import pickle
import sys

from sertit import arcpy, ci

ci.reduce_verbosity()


# flake8: noqa
def test_arcpy():
    """Test CI functions"""
    arcpy.init_conda_arcpy_env()
    import geopandas as gpd
    import rasterio
    from lxml import etree  # Should work


def test_create_conda_env_cli(tmp_path):
    """Test create_conda_env_cli function"""

    # Create a dumb fct
    def my_tool_core(a):
        return a + 2, "test"

    # Pickle paths
    in_pkl_path = tmp_path / "input.pkl"
    out_pkl_path = tmp_path / "output.pkl"

    # Add input for the fct in a pickle
    with open(in_pkl_path, "wb") as fp_in:
        pickle.dump({"a": 1}, fp_in)

    # Create CLI and run fct
    sys.argv = [__file__, "-i", str(in_pkl_path), "-o", str(out_pkl_path)]
    arcpy.create_conda_env_cli(standalone_mode=False)(my_tool_core)

    # Load output
    with open(out_pkl_path, "rb") as fp_out:
        out = pickle.load(fp_out)

    # Check output
    ci.assert_val(out, (3, "test"), "Test create_backend_cli")



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unistra tools
"""

import configparser
import logging
import os
from contextlib import contextmanager
from pathlib import Path

from sertit import AnyPath, s3
from sertit.logs import SU_NAME
from sertit.s3 import USE_S3_STORAGE, temp_s3
from sertit.types import AnyPathType

LOGGER = logging.getLogger(SU_NAME)

UNISTRA_S3_ENDPOINT = "s3.unistra.fr"
"""
Unistra S3 compatible storage endpoint: s3.unistra.fr
"""

UNISTRA_S3_ENPOINT = UNISTRA_S3_ENDPOINT
# Legacy, to be removed in v2.0


def s3_env(*args, **kwargs):
    """
    Create Unistra's S3 compatible storage environment.

    This function searches for S3 configuration in many places.
    It does apply configuration variables precedence, and you might have a use for it.
    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    #. AWS profile
    #. AWS environment variable

    Profile unistra is first read from X:/SI/Secrets/config and X:/SI/Secrets/credentials.
    If this file does not exist, it fallbacks to local file $USER/.aws/config and $USER/.aws/credentials.

    You can use ready-to-use environements provided by the Sertit or asks for s3 credentials.

    Args:
        function (Callable): Function to decorate

    Returns:
        Callable: decorated function

    Example:
        >>> from sertit.unistra import s3_env
        >>> from sertit import AnyPath
        >>> @s3_env
        >>> def file_exists(path: str):
        >>>     pth = AnyPath(path)
        >>>     print(pth.exists())
        >>> file_exists("s3://sertit-geodatastore/GLOBAL/COPDEM_30m/COPDEM_30m.vrt")
        True
    """
    _set_aws_file_path()
    use_s3 = kwargs.pop("use_s3_env_var", USE_S3_STORAGE)
    extra_args = {"profile_name": "unistra"} if does_unistra_profile_exist() else {}
    extra_args["endpoint"] = UNISTRA_S3_ENDPOINT
    return s3.s3_env(use_s3_env_var=use_s3, **extra_args)(*args, **kwargs)


@contextmanager
def unistra_s3() -> None:
    """
    Initialize a temporary S3 environment as a context manager, with Unistra endpoint

    This function searches for S3 configuration in many places.
    It does apply configuration variables precedence, and you might have a use for it.
    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    #. AWS profile "unistra"
    #. AWS environment variable

    Profile unistra is first read from X:/SI/Secrets/config and X:/SI/Secrets/credentials.
    If this file does not exist, it fallbacks to local file $USER/.aws/config and $USER/.aws/credentials.

    You can use ready-to-use environements provided by the Sertit or asks for s3 credentials.

    Args:
        default_endpoint (str):Default Endpoint to look for

    Example:
        >>> from sertit.unistra import unistra_s3
        >>> from sertit import AnyPath
        >>> def file_exists(path: str):
        >>>     with unistra_s3():
        >>>         pth = AnyPath(path)
        >>>         print(pth.exists())
        >>> file_exists("s3://sertit-geodatastore/GLOBAL/COPDEM_30m/COPDEM_30m.vrt")
        True
    """
    _set_aws_file_path()
    try:
        extra_args = {"profile_name": "unistra"} if does_unistra_profile_exist() else {}
        extra_args["endpoint"] = UNISTRA_S3_ENDPOINT
        with temp_s3(**extra_args):
            yield
    finally:
        pass


def define_s3_client():
    """
    Define Unistra's S3 client

    This function searches for S3 configuration in many places.
    It does apply configuration variables precedence, and you might have a use for it.
    Here is the order of precedence from least to greatest
    (the last listed configuration variables override all other variables):

    #. AWS profile
    #. AWS environment variable

    Profile unistra is first read from X:/SI/Secrets/config and X:/SI/Secrets/credentials.
    If this file does not exist, it fallbacks to local file $USER/.aws/config and $USER/.aws/credentials.

    You can use ready-to-use environements provided by the Sertit or asks for s3 credentials.

    """
    _set_aws_file_path()
    profile_arg = {"profile_name": "unistra"} if does_unistra_profile_exist() else {}
    return s3.define_s3_client(endpoint=UNISTRA_S3_ENDPOINT, **profile_arg)


def get_geodatastore() -> AnyPathType:
    """
    Get database directory.

    If :any:`USE_S3_STORAGE` is set to ``1``, this function returns ``AnyPath("s3://sertit-geodatastore")``.

    If :any:`USE_S3_STORAGE` is set to ``0`` it returns:

    * ``AnyPath("//ds2/database02/BASES_DE_DONNEES")`` if code is running on Windows
    * ``AnyPath(/home/ds2_db2/BASE_DE_DONNESS`)`` if code is running in a Docker containers on Windows

    Returns:
        AnyPath: Database directory

    Example:
        Don't set manually ``USE_S3_STORAGE`` with ``os.environ`` !

        >>> from sertit.unistra import get_geodatastore
        >>> import os
        >>> os.environ["USE_S3_STORAGE"] = "1"
        >>> print(get_geodatastore())
        s3://sertit-geodatastore

        >>> from sertit.unistra import get_geodatastore
        >>> import os
        >>> os.environ["USE_S3_STORAGE"] = "0"
        >>> print(get_geodatastore())
        //ds2/database02/BASES_DE_DONNEES/GLOBAL
    """
    if int(os.getenv(s3.USE_S3_STORAGE, 0)):
        # Define S3 client for S3 paths
        define_s3_client()
        return AnyPath("s3://sertit-geodatastore")
    else:
        try:
            db_dir = AnyPath(get_db2_path(), "BASES_DE_DONNEES")
        except NotADirectoryError:  # pragma: no cover
            db_dir = AnyPath("/home", "ds2_db2", "BASES_DE_DONNEES")

        if not db_dir.is_dir():  # pragma: no cover
            raise NotADirectoryError("Impossible to open database directory !")

    return AnyPath(db_dir)


def does_unistra_profile_exist() -> bool:
    """
    This function checks if "unistra" AWS profile exists.

    Returns: True if "unistra" AWS profile is present on the machine, False otherwise

    """
    config = configparser.ConfigParser()
    default_credentials_path = Path.home() / ".aws" / "credentials"
    credentials_path = os.getenv(
        "AWS_SHARED_CREDENTIALS_FILE", str(default_credentials_path)
    )
    credentials_path = AnyPath(credentials_path)
    if not credentials_path.exists():
        return False
    else:  # pragma: no cover
        config.read(credentials_path)
        return "unistra" in config.sections()


def _set_aws_file_path() -> None:
    """
    This function sets AWS filepath config and credentials inside X network drive "X:/SI/Secrets/AWS"
    Returns:

    """
    config_file = AnyPath("X:") / "SI" / "Secrets" / "AWS" / "config"
    credentials_file = AnyPath("X:") / "SI" / "Secrets" / "AWS" / "credentials"
    if config_file.exists():
        os.environ["AWS_CONFIG_FILE"] = str(config_file)
    if credentials_file.exists():
        os.environ["AWS_SHARED_CREDENTIALS_FILE"] = str(credentials_file)


def get_mnt_path() -> str:
    """
    Return mounting directory :code:`/mnt`.

    Warnings:
        This won't work on Windows !

    Returns:
        str: Mounting directory

    Example:
        >>> get_mnt_path()
        '/mnt'
    """
    return r"/mnt"


def _get_db_path(db_nb=2) -> str:
    """
    Returns DSx database0x path

    - :code:`/mnt/ds2_dbx` when mounted (docker...)
    """
    db_path = f"{get_mnt_path()}/ds2_db{db_nb}"

    if not os.path.isdir(db_path):  # pragma: no cover
        raise NotADirectoryError(f"Impossible to open {db_path}!")
    return db_path


def get_db2_path() -> str:
    """
    Returns DS2 database02 path

    - :code:`/mnt/ds2_db2` when mounted (docker...)

    Returns:
        str: Mounted directory

    Example:
        >>> get_db2_path()
        '/mnt/ds2_db2'
    """
    return _get_db_path(2)


def get_db3_path() -> str:
    """
    Returns DS2 database03 path

    - :code:`/mnt/ds2_db3` when mounted (docker...)

    Returns:
        str: Mounted directory

    Example:
        >>> get_db3_path()
        '/mnt/ds2_db3'
    """
    return _get_db_path(3)



import logging
import re
from functools import cmp_to_key
from pathlib import Path

from sertit import strings
from sertit.logs import SU_NAME

LOGGER = logging.getLogger(SU_NAME)


def _compare_str_containing_version(str1: str, str2: str):
    """
    Compare two strings containing a software version, for example "qgis 3.34.5" and "qgis 3.35.6".

    Args:
        str1:
        str2:

    Returns:
        Let's say str1 and str2 contains respectively the versions v1 and v2. So, this function returns:
        -1 if v1 <= v2
        0 if v1 == v2
        1 if v1 >= v2

    """
    from semver import Version

    match_pattern = r"\d+\.\d+\.?\d+?"
    v1 = Version.parse(
        re.search(match_pattern, str1).group(), optional_minor_and_patch=True
    )
    v2 = Version.parse(
        re.search(match_pattern, str2).group(), optional_minor_and_patch=True
    )
    if v1 == v2:
        return 0
    elif v1 < v2:
        return -1
    else:
        return 1


def qgis_bin() -> Path:
    """
    Looking for qgis bin directory in Windows filesystem and return it.

    Returns: None if qgis bin directory is not found or a path to the qgis bin directory otherwise.
    """
    # Find QGis in ProgramFiles
    qgis_dirs = []
    parent_path_qgis = Path("C:/") / "Program Files"
    if parent_path_qgis.exists():  # pragma: no cover
        qgis_dirs = [
            str(pt)
            for pt in parent_path_qgis.iterdir()
            if pt.stem.lower().startswith("qgis")
        ]

    if len(qgis_dirs) > 0:  # pragma: no cover
        # Take the latest one available
        qgis_dirs = sorted(qgis_dirs, key=cmp_to_key(_compare_str_containing_version))
        qgis_valid_dirs = []
        for path in qgis_dirs:
            bin_path = Path(path) / "bin"
            if bin_path.exists():
                qgis_valid_dirs.append(bin_path)

        qgis_dir = None if len(qgis_valid_dirs) == 0 else qgis_valid_dirs[-1]

    # Find QGis in osgeo4w
    else:
        osgeo_path = Path("C:/") / "osgeo4w" / "bin"
        qgis_dir = osgeo_path if osgeo_path.exists() else None

    return qgis_dir


def gdalbuildvrt_exe() -> str:
    """
    Looking for gdalbuildvrt exe from the path or inside qgis bin directory.

    Returns:
        str: gdalbuildvrt exe

    """
    qgis = qgis_bin()

    if qgis is not None:  # pragma: no cover
        gdal_build_vrt_exe = strings.to_cmd_string(str(qgis / "gdalbuildvrt.exe"))
    else:
        gdal_build_vrt_exe = "gdalbuildvrt"

    return gdal_build_vrt_exe



# Copyright 2025, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of sertit-utils project
#     https://github.com/sertit/sertit-utils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Script testing dask functions"""

import logging
import os

import pytest
import tempenv

from ci.script_utils import rasters_path, s3_env
from sertit import ci, dask, rasters
from sertit.dask import SERTIT_DEFAULT_CHUNKS, get_default_chunks
from sertit.logs import SU_NAME

LOGGER = logging.getLogger(SU_NAME)

ci.reduce_verbosity()


def test_default_chunks():
    with tempenv.TemporaryEnvironment({SERTIT_DEFAULT_CHUNKS: None}):
        ci.assert_val(get_default_chunks(), "auto", "default chunks: unset")

    with tempenv.TemporaryEnvironment({SERTIT_DEFAULT_CHUNKS: "none"}):
        ci.assert_val(get_default_chunks(), None, "default chunks: none")

    with tempenv.TemporaryEnvironment({SERTIT_DEFAULT_CHUNKS: "auto"}):
        ci.assert_val(get_default_chunks(), "auto", "default chunks: auto")

    with tempenv.TemporaryEnvironment({SERTIT_DEFAULT_CHUNKS: "true"}):
        ci.assert_val(get_default_chunks(), True, "default chunks: true")


@s3_env
def test_computed_lazy():
    """Test read function"""
    raster_path = rasters_path().joinpath("raster.tif")

    # Read already computed
    np_arr = rasters.read(raster_path, chunks=None)
    assert np_arr is not None
    assert not dask.is_chunked(np_arr)
    ci.assert_computed(np_arr)

    # Read lazy
    da_arr = rasters.read(raster_path)
    assert da_arr is not None
    assert dask.is_chunked(da_arr)
    ci.assert_chunked(da_arr)

    # Assert computed after compute
    ci.assert_computed(da_arr.compute())


def test_raise_compute():
    raster_path = rasters_path().joinpath("raster.tif")

    # Read lazily
    with dask.raise_if_dask_computes():
        da = rasters.read(raster_path)

    # Ensure computing raises
    with (
        pytest.raises(RuntimeError),
        dask.raise_if_dask_computes(force_synchronous=True),
    ):
        da.compute()

    # Ensure computing raises
    with (
        pytest.raises(RuntimeError),
        dask.raise_if_dask_computes(force_synchronous=True, max_computes=1),
    ):
        da.compute()
        da.compute()

    # Ensure computing raises
    with (
        pytest.raises(RuntimeError),
        dask.raise_if_dask_computes(force_synchronous=True, nof_computes=1),
    ):
        da.compute()
        da.compute()

    # Ensure computing raises
    with (
        pytest.raises(RuntimeError),
        dask.raise_if_dask_computes(force_synchronous=True, nof_computes=3),
    ):
        da.compute()

    # Ensure computing don't raise
    with dask.raise_if_dask_computes(force_synchronous=True, nof_computes=1):
        da.compute()

    # Ensure computing don't raise
    with dask.raise_if_dask_computes(force_synchronous=True, max_computes=1):
        da.compute()

    # Ensure computing don't raise id specified
    with dask.raise_if_dask_computes(force_synchronous=True, dont_raise=True):
        da.compute()

    # Check if da is still chunked
    ci.assert_chunked(da)

    # First gotcha
    with (
        pytest.raises(RuntimeError),
        dask.raise_if_dask_computes(force_synchronous=True),
    ):
        if da.max() > 5:
            print("Gotcha!")

    # Check if da is still chunked
    ci.assert_chunked(da)

    # Second gotchas
    # with pytest.raises(RuntimeError), dask.raise_if_dask_computes():
    #     pass

    # Check if da is still chunked
    # ci.assert_chunked(da)

    # Other gotchas from digitize: TODO?
    # import dask.array as da
    # import xarray as xr
    # import numpy as np
    # with pytest.raises(RuntimeError), dask.raise_if_dask_computes():
    #     values = da.from_array([1, 2, 3], chunks="auto")
    #     arr = values.vindex[
    #         xr.apply_ufunc(np.digitize, a, [0.1, 0.2], dask="allowed").data
    #     ]


def test_env_var(capfd):
    with dask.get_or_create_dask_client(
        processes=True, env_vars={"PROCESSES": "True"}
    ) as client:
        client.run(lambda: print("1. processes=" + os.environ["PROCESSES"]))
    out, err = capfd.readouterr()
    assert "1. processes=True" in out
    assert "1. processes=False" not in out

    with dask.get_or_create_dask_client(
        processes=False, env_vars={"PROCESSES": "False"}
    ) as client:
        client.run(lambda: print("2. processes=" + os.environ["PROCESSES"]))
    out, err = capfd.readouterr()
    assert "2. processes=False" in out
    assert "2. processes=True" not in out
